#!/usr/bin/python
# Copyright (c) 2016 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittests for resamplers."""

import os
import shutil
import tempfile
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyproj import CRS

from satpy.resample.native import NativeResampler
from satpy.utils import PerformanceWarning


def get_test_data(input_shape=(100, 50), output_shape=(200, 100), output_proj=None,
                  input_dims=("y", "x")):
    """Get common data objects used in testing.

    Returns:
        tuple:

        * input_data_on_area: DataArray with dimensions as if it is a gridded
          dataset.
        * input_area_def: AreaDefinition of the above DataArray
        * input_data_on_swath: DataArray with dimensions as if it is a swath.
        * input_swath: SwathDefinition of the above DataArray
        * target_area_def: AreaDefinition to be used as a target for resampling

    """
    import dask.array as da
    from pyresample.geometry import AreaDefinition, SwathDefinition
    from xarray import DataArray
    ds1 = DataArray(da.zeros(input_shape, chunks=85),
                    dims=input_dims,
                    attrs={"name": "test_data_name", "test": "test"})
    if input_dims and "y" in input_dims:
        ds1 = ds1.assign_coords(y=da.arange(input_shape[-2], chunks=85))
    if input_dims and "x" in input_dims:
        ds1 = ds1.assign_coords(x=da.arange(input_shape[-1], chunks=85))
    if input_dims and "bands" in input_dims:
        ds1 = ds1.assign_coords(bands=list("RGBA"[:ds1.sizes["bands"]]))

    input_proj_str = ("+proj=geos +lon_0=-95.0 +h=35786023.0 +a=6378137.0 "
                      "+b=6356752.31414 +sweep=x +units=m +no_defs")
    crs = CRS(input_proj_str)
    source = AreaDefinition(
        "test_target",
        "test_target",
        "test_target",
        crs,
        input_shape[1],  # width
        input_shape[0],  # height
        (-1000., -1500., 1000., 1500.))
    ds1.attrs["area"] = source
    ds1 = ds1.assign_coords(crs=crs)

    ds2 = ds1.copy()
    input_area_shape = tuple(ds1.sizes[dim] for dim in ds1.dims
                             if dim in ["y", "x"])
    geo_dims = ("y", "x") if input_dims else None
    lons = da.random.random(input_area_shape, chunks=50)
    lats = da.random.random(input_area_shape, chunks=50)
    swath_def = SwathDefinition(
        DataArray(lons, dims=geo_dims),
        DataArray(lats, dims=geo_dims))
    ds2.attrs["area"] = swath_def
    crs = CRS.from_string("+proj=latlong +datum=WGS84 +ellps=WGS84")
    ds2 = ds2.assign_coords(crs=crs)

    # set up target definition
    output_proj_str = ("+proj=lcc +datum=WGS84 +ellps=WGS84 "
                       "+lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs")
    output_proj_str = output_proj or output_proj_str
    target = AreaDefinition(
        "test_target",
        "test_target",
        "test_target",
        CRS(output_proj_str),
        output_shape[1],  # width
        output_shape[0],  # height
        (-1000., -1500., 1000., 1500.),
    )
    return ds1, source, ds2, swath_def, target


class TestHLResample(unittest.TestCase):
    """Test the higher level resampling functions."""

    def test_type_preserve(self):
        """Check that the type of resampled datasets is preserved."""
        from pyresample.geometry import SwathDefinition

        from satpy.resample.base import resample_dataset
        source_area = SwathDefinition(xr.DataArray(da.arange(4, chunks=5).reshape((2, 2)), dims=["y", "x"]),
                                      xr.DataArray(da.arange(4, chunks=5).reshape((2, 2)), dims=["y", "x"]))
        dest_area = SwathDefinition(xr.DataArray(da.arange(4, chunks=5).reshape((2, 2)) + .0001, dims=["y", "x"]),
                                    xr.DataArray(da.arange(4, chunks=5).reshape((2, 2)) + .0001, dims=["y", "x"]))
        expected_gap = np.array([[1, 2], [3, 255]])
        data = xr.DataArray(da.from_array(expected_gap, chunks=5), dims=["y", "x"])
        data.attrs["_FillValue"] = 255
        data.attrs["area"] = source_area
        res = resample_dataset(data, dest_area)
        assert res.dtype == data.dtype
        assert np.all(res.values == expected_gap)

        expected_filled = np.array([[1, 2], [3, 3]])
        res = resample_dataset(data, dest_area, radius_of_influence=1000000)
        assert res.dtype == data.dtype
        assert np.all(res.values == expected_filled)


class TestKDTreeResampler(unittest.TestCase):
    """Test the kd-tree resampler."""

    @mock.patch("satpy.resample.kdtree.xr.Dataset")
    @mock.patch("satpy.resample.kdtree.zarr.open")
    @mock.patch("satpy.resample.kdtree.KDTreeResampler._create_cache_filename")
    @mock.patch("pyresample.kd_tree.XArrayResamplerNN")
    def test_kd_resampling(self, xr_resampler, create_filename, zarr_open,
                           xr_dset):
        """Test the kd resampler."""
        from satpy.resample.kdtree import KDTreeResampler
        data, source_area, swath_data, source_swath, target_area = get_test_data()
        mock_dset = mock.MagicMock()
        xr_dset.return_value = mock_dset
        resampler = KDTreeResampler(source_swath, target_area)
        resampler.precompute(
            mask=da.arange(5, chunks=5).astype(bool), cache_dir=".")
        xr_resampler.assert_called_once()
        resampler.resampler.get_neighbour_info.assert_called()
        # swath definitions should not be cached
        assert len(mock_dset.to_zarr.mock_calls) == 0
        resampler.resampler.reset_mock()

        resampler = KDTreeResampler(source_area, target_area)
        resampler.precompute()
        resampler.resampler.get_neighbour_info.assert_called_with(mask=None)

        try:
            the_dir = tempfile.mkdtemp()
            resampler = KDTreeResampler(source_area, target_area)
            create_filename.return_value = os.path.join(the_dir, "test_cache.zarr")
            zarr_open.side_effect = ValueError()
            resampler.precompute(cache_dir=the_dir)
            # assert data was saved to the on-disk cache
            assert len(mock_dset.to_zarr.mock_calls) == 1
            # assert that zarr_open was called to try to zarr_open something from disk
            assert len(zarr_open.mock_calls) == 1
            # we should have cached things in-memory
            assert len(resampler._index_caches) == 1
            nbcalls = len(resampler.resampler.get_neighbour_info.mock_calls)
            # test reusing the resampler
            zarr_open.side_effect = None
            # The kdtree shouldn't be available after saving cache to disk
            assert resampler.resampler.delayed_kdtree is None

            class FakeZarr(dict):

                def close(self):
                    pass

                def astype(self, dtype):
                    pass

            zarr_open.return_value = FakeZarr(valid_input_index=1,
                                              valid_output_index=2,
                                              index_array=3,
                                              distance_array=4)
            resampler.precompute(cache_dir=the_dir)
            # we already have things cached in-memory, no need to save again
            assert len(mock_dset.to_zarr.mock_calls) == 1
            # we already have things cached in-memory, don't need to load
            assert len(zarr_open.mock_calls) == 1
            # we should have cached things in-memory
            assert len(resampler._index_caches) == 1
            assert len(resampler.resampler.get_neighbour_info.mock_calls) == nbcalls

            # test loading saved resampler
            resampler = KDTreeResampler(source_area, target_area)
            resampler.precompute(cache_dir=the_dir)
            assert len(zarr_open.mock_calls) == 4
            assert len(resampler.resampler.get_neighbour_info.mock_calls) == nbcalls
            # we should have cached things in-memory now
            assert len(resampler._index_caches) == 1
        finally:
            shutil.rmtree(the_dir)

        fill_value = 8
        resampler.compute(data, fill_value=fill_value)
        resampler.resampler.get_sample_from_neighbour_info.assert_called_with(data, fill_value)


class TestNativeResampler:
    """Tests for the 'native' resampling method."""

    def setup_method(self):
        """Create test data used by multiple tests."""
        self.d_arr = da.zeros((6, 20), chunks=4)

    def test_expand_reduce_replicate(self):
        """Test classmethod 'expand_reduce' to replicate by 2."""
        new_data = NativeResampler._expand_reduce(self.d_arr, {0: 2., 1: 2.})
        assert new_data.shape == (12, 40)

    def test_expand_reduce_aggregate(self):
        """Test classmethod 'expand_reduce' to aggregate by half."""
        new_data = NativeResampler._expand_reduce(self.d_arr, {0: .5, 1: .5})
        assert new_data.shape == (3, 10)

    def test_expand_reduce_aggregate_identity(self):
        """Test classmethod 'expand_reduce' returns the original dask array when factor is 1."""
        new_data = NativeResampler._expand_reduce(self.d_arr, {0: 1., 1: 1.})
        assert new_data.shape == (6, 20)
        assert new_data is self.d_arr

    @pytest.mark.parametrize("dim0_factor", [1. / 4, 0.333323423, 1.333323423])
    def test_expand_reduce_aggregate_invalid(self, dim0_factor):
        """Test classmethod 'expand_reduce' fails when factor does not divide evenly."""
        with pytest.raises(ValueError, match="[Aggregation, Expand] .*"):
            NativeResampler._expand_reduce(self.d_arr, {0: dim0_factor, 1: 1.})

    def test_expand_reduce_agg_rechunk(self):
        """Test that an incompatible factor for the chunk size is rechunked.

        This can happen when a user chunks their data that makes sense for
        the overall shape of the array and for their local machine's
        performance, but the resulting resampling factor does not divide evenly
        into that chunk size.

        """
        d_arr = da.zeros((6, 20), chunks=3)
        text = "Array chunk size is not divisible by aggregation factor. Re-chunking to continue native resampling."
        with pytest.warns(PerformanceWarning, match=text):
            new_data = NativeResampler._expand_reduce(d_arr, {0: 0.5, 1: 0.5})
        assert new_data.shape == (3, 10)

    def test_expand_reduce_numpy(self):
        """Test classmethod 'expand_reduce' converts numpy arrays to dask arrays."""
        n_arr = np.zeros((6, 20))
        new_data = NativeResampler._expand_reduce(n_arr, {0: 2., 1: 1.0})
        np.testing.assert_equal(new_data.compute()[::2, :], n_arr)

    def test_expand_dims(self):
        """Test expanding native resampling with 2D data."""
        ds1, source_area, _, _, target_area = get_test_data()
        # source geo def doesn't actually matter
        resampler = NativeResampler(source_area, target_area)
        new_data = resampler.resample(ds1)
        assert new_data.shape == (200, 100)
        new_data2 = resampler.resample(ds1.compute())
        np.testing.assert_equal(new_data.compute().data, new_data2.compute().data)
        assert "y" in new_data.coords
        assert "x" in new_data.coords
        assert "crs" in new_data.coords
        assert isinstance(new_data.coords["crs"].item(), CRS)
        assert "lambert" in new_data.coords["crs"].item().coordinate_operation.method_name.lower()
        assert new_data.coords["y"].attrs["units"] == "meter"
        assert new_data.coords["x"].attrs["units"] == "meter"
        assert target_area.crs == new_data.coords["crs"].item()

    def test_expand_dims_3d(self):
        """Test expanding native resampling with 3D data."""
        ds1, source_area, _, _, target_area = get_test_data(
            input_shape=(3, 100, 50), input_dims=("bands", "y", "x"))
        # source geo def doesn't actually matter
        resampler = NativeResampler(source_area, target_area)
        new_data = resampler.resample(ds1)
        assert new_data.shape == (3, 200, 100)
        new_data2 = resampler.resample(ds1.compute())
        np.testing.assert_equal(new_data.compute().data, new_data2.compute().data)
        assert "y" in new_data.coords
        assert "x" in new_data.coords
        assert "bands" in new_data.coords
        np.testing.assert_equal(new_data.coords["bands"].values, ["R", "G", "B"])
        assert "crs" in new_data.coords
        assert isinstance(new_data.coords["crs"].item(), CRS)
        assert "lambert" in new_data.coords["crs"].item().coordinate_operation.method_name.lower()
        assert new_data.coords["y"].attrs["units"] == "meter"
        assert new_data.coords["x"].attrs["units"] == "meter"
        assert target_area.crs == new_data.coords["crs"].item()

    def test_expand_without_dims(self):
        """Test expanding native resampling with no dimensions specified."""
        ds1, source_area, _, _, target_area = get_test_data(input_dims=None)
        # source geo def doesn't actually matter
        resampler = NativeResampler(source_area, target_area)
        new_data = resampler.resample(ds1)
        assert new_data.shape == (200, 100)
        new_data2 = resampler.resample(ds1.compute())
        np.testing.assert_equal(new_data.compute().data, new_data2.compute().data)
        assert "crs" in new_data.coords
        assert isinstance(new_data.coords["crs"].item(), CRS)
        assert "lambert" in new_data.coords["crs"].item().coordinate_operation.method_name.lower()
        assert target_area.crs == new_data.coords["crs"].item()

    def test_expand_without_dims_4D(self):
        """Test expanding native resampling with 4D data with no dimensions specified."""
        ds1, source_area, _, _, target_area = get_test_data(
            input_shape=(2, 3, 100, 50), input_dims=None)
        # source geo def doesn't actually matter
        resampler = NativeResampler(source_area, target_area)
        with pytest.raises(ValueError, match="Can only handle 2D or 3D arrays without dimensions."):
            resampler.resample(ds1)

    def test_reduce_first_dim_rechunk_second_dim_good(self):
        """Test inconsistent chunks are handled consistently."""
        # See https://github.com/pytroll/satpy/issues/3304
        row_chunks = (6, 6, 7, 6, 6, 7, 8, 6, 4) * 5
        col_chunks = (280,)
        d_arr = da.zeros((280, 280), chunks=(row_chunks, col_chunks))
        text = "Array chunk size is not divisible by aggregation factor. Re-chunking to continue native resampling."
        with pytest.warns(PerformanceWarning, match=text):
            new_data = NativeResampler._expand_reduce(d_arr, {0: 0.5, 1: 0.5})
        assert new_data.shape == (140, 140)


class TestBilinearResampler(unittest.TestCase):
    """Test the bilinear resampler."""

    @mock.patch("satpy.resample.kdtree._move_existing_caches")
    @mock.patch("satpy.resample.kdtree.BilinearResampler._create_cache_filename")
    @mock.patch("pyresample.bilinear.XArrayBilinearResampler")
    def test_bil_resampling(self, xr_resampler, create_filename,
                            move_existing_caches):
        """Test the bilinear resampler."""
        from satpy.resample.kdtree import BilinearResampler
        data, source_area, swath_data, source_swath, target_area = get_test_data()

        # Test that bilinear resampling info calculation is called
        resampler = BilinearResampler(source_swath, target_area)
        resampler.precompute(
            mask=da.arange(5, chunks=5).astype(bool))
        resampler.resampler.load_resampling_info.assert_not_called()
        resampler.resampler.get_bil_info.assert_called_once()
        resampler.resampler.reset_mock()

        # Test that get_sample_from_bil_info is called properly
        fill_value = 8
        resampler.resampler.get_sample_from_bil_info.return_value = \
            xr.DataArray(da.zeros(target_area.shape), dims=("y", "x"))
        new_data = resampler.compute(data, fill_value=fill_value)
        resampler.resampler.get_sample_from_bil_info.assert_called_with(
            data, fill_value=fill_value, output_shape=target_area.shape)
        assert "y" in new_data.coords
        assert "x" in new_data.coords
        assert "crs" in new_data.coords
        assert isinstance(new_data.coords["crs"].item(), CRS)
        assert "lambert" in new_data.coords["crs"].item().coordinate_operation.method_name.lower()
        assert new_data.coords["y"].attrs["units"] == "meter"
        assert new_data.coords["x"].attrs["units"] == "meter"
        assert target_area.crs == new_data.coords["crs"].item()

        # Test that the resampling info is tried to read from the disk
        resampler = BilinearResampler(source_swath, target_area)
        resampler.precompute(cache_dir=".")
        resampler.resampler.load_resampling_info.assert_called()

        # Test caching the resampling info
        try:
            the_dir = tempfile.mkdtemp()
            resampler = BilinearResampler(source_area, target_area)
            create_filename.return_value = os.path.join(the_dir, "test_cache.zarr")
            xr_resampler.return_value.load_resampling_info.side_effect = IOError

            resampler.precompute(cache_dir=the_dir)
            resampler.resampler.save_resampling_info.assert_called()
            # assert data was saved to the on-disk cache
            resampler.resampler.save_resampling_info.assert_called_once()

            nbcalls = resampler.resampler.get_bil_info.call_count
            resampler.resampler.load_resampling_info.side_effect = None

            resampler.precompute(cache_dir=the_dir)
            # we already have things cached in-memory, no need to save again
            resampler.resampler.save_resampling_info.assert_called_once()
            # we already have things cached in-memory, don't need to load
            assert resampler.resampler.get_bil_info.call_count == nbcalls

            # test loading saved resampler
            resampler = BilinearResampler(source_area, target_area)
            resampler.precompute(cache_dir=the_dir)
            assert resampler.resampler.load_resampling_info.call_count == 3
            assert resampler.resampler.get_bil_info.call_count == nbcalls

            resampler = BilinearResampler(source_area, target_area)
            resampler.precompute(cache_dir=the_dir)
            resampler.save_bil_info(cache_dir=the_dir)
            zarr_file = os.path.join(the_dir, "test_cache.zarr")
            # Save again faking the cache file already exists
            with mock.patch("os.path.exists") as exists:
                exists.return_value = True
                resampler.save_bil_info(cache_dir=the_dir)
            move_existing_caches.assert_called_once_with(the_dir, zarr_file)

        finally:
            shutil.rmtree(the_dir)

    def test_move_existing_caches(self):
        """Test that existing caches are moved to a subdirectory."""
        try:
            the_dir = tempfile.mkdtemp()
            # Test that existing cache file is moved away
            zarr_file = os.path.join(the_dir, "test.zarr")
            with open(zarr_file, "w") as fid:
                fid.write("42")
            from satpy.resample.kdtree import _move_existing_caches
            _move_existing_caches(the_dir, zarr_file)
            assert not os.path.exists(zarr_file)
            assert os.path.exists(os.path.join(the_dir, "moved_by_satpy", "test.zarr"))
            # Run again to see that the existing dir doesn't matter
            with open(zarr_file, "w") as fid:
                fid.write("42")
            _move_existing_caches(the_dir, zarr_file)
        finally:
            shutil.rmtree(the_dir)


class TestBucketAvg(unittest.TestCase):
    """Test the bucket resampler."""

    def setUp(self):
        """Create fake area definitions and resampler to be tested."""
        from satpy.resample.bucket import BucketAvg
        get_lonlats = mock.MagicMock()
        get_lonlats.return_value = (1, 2)
        get_proj_vectors = mock.MagicMock()
        get_proj_vectors.return_value = ([1, 2, 3, 4, 5],  [1, 2, 3, 4, 5])
        self.source_geo_def = mock.MagicMock(get_lonlats=get_lonlats)
        self.target_geo_def = mock.MagicMock(get_lonlats=get_lonlats, crs=None, get_proj_vectors=get_proj_vectors)
        self.bucket = BucketAvg(self.source_geo_def, self.target_geo_def)

    def test_init(self):
        """Test bucket resampler initialization."""
        assert self.bucket.resampler is None
        assert self.bucket.source_geo_def == self.source_geo_def
        assert self.bucket.target_geo_def == self.target_geo_def

    @mock.patch("pyresample.bucket.BucketResampler")
    def test_precompute(self, bucket):
        """Test bucket resampler precomputation."""
        bucket.return_value = True
        self.bucket.precompute()
        assert self.bucket.resampler
        bucket.assert_called_once_with(self.target_geo_def, 1, 2)

    def _compute_mocked_bucket_avg(self, data, return_data=None, **kwargs):
        """Compute the mocked bucket average."""
        self.bucket.resampler = mock.MagicMock()
        if return_data is not None:
            self.bucket.resampler.get_average.return_value = return_data
        else:
            self.bucket.resampler.get_average.return_value = data
        res = self.bucket.compute(data, **kwargs)
        return res

    def test_compute(self):
        """Test bucket resampler computation."""
        # 1D data
        data = da.ones((5,))
        res = self._compute_mocked_bucket_avg(data, fill_value=2)
        assert res.shape == (1, 5)
        # 2D data
        data = da.ones((5, 5))
        res = self._compute_mocked_bucket_avg(data, fill_value=2)
        assert res.shape == (1, 5, 5)
        # 3D data
        data = da.ones((3, 5, 5))
        self.bucket.resampler.get_average.return_value = data[0, :, :]
        res = self._compute_mocked_bucket_avg(data, return_data=data[0, :, :], fill_value=2)
        assert res.shape == (3, 5, 5)

    def test_compute_and_use_skipna_handling(self):
        """Test bucket resampler computation and use skipna handling."""
        data = da.ones((5,))

        self._compute_mocked_bucket_avg(data, fill_value=2, skipna=False)
        self.bucket.resampler.get_average.assert_called_once_with(
            data,
            fill_value=2,
            skipna=False)

        self._compute_mocked_bucket_avg(data, fill_value=2)
        self.bucket.resampler.get_average.assert_called_once_with(
            data,
            fill_value=2,
            skipna=True)

    @mock.patch("pyresample.bucket.BucketResampler")
    def test_resample(self, pyresample_bucket):
        """Test bucket resamplers resample method."""
        self.bucket.resampler = mock.MagicMock()
        self.bucket.precompute = mock.MagicMock()
        self.bucket.compute = mock.MagicMock()

        # 1D input data
        data = xr.DataArray(da.ones((5,)), dims=("foo"), attrs={"bar": "baz"})
        self.bucket.compute.return_value = da.ones((5, 5))
        res = self.bucket.resample(data)
        self.bucket.precompute.assert_called_once()
        self.bucket.compute.assert_called_once()
        assert res.shape == (5, 5)
        assert res.dims == ("y", "x")
        assert "bar" in res.attrs
        assert res.attrs["bar"] == "baz"

        # 2D input data
        data = xr.DataArray(da.ones((5, 5)), dims=("foo", "bar"))
        self.bucket.compute.return_value = da.ones((5, 5))
        res = self.bucket.resample(data)
        assert res.shape == (5, 5)
        assert res.dims == ("y", "x")

        # 3D input data with 'bands' dim
        data = xr.DataArray(da.ones((1, 5, 5)), dims=("bands", "foo", "bar"),
                            coords={"bands": ["L"]})
        self.bucket.compute.return_value = da.ones((1, 5, 5))
        res = self.bucket.resample(data)
        assert res.shape == (1, 5, 5)
        assert res.dims == ("bands", "y", "x")
        assert res.coords["bands"] == ["L"]

        # 3D input data with misc dim names
        data = xr.DataArray(da.ones((3, 5, 5)), dims=("foo", "bar", "baz"))
        self.bucket.compute.return_value = da.ones((3, 5, 5))
        res = self.bucket.resample(data)
        assert res.shape == (3, 5, 5)
        assert res.dims == ("foo", "bar", "baz")


class TestBucketSum(unittest.TestCase):
    """Test the sum bucket resampler."""

    def setUp(self):
        """Create fake area definitions and resampler to be tested."""
        from satpy.resample.bucket import BucketSum
        get_lonlats = mock.MagicMock()
        get_lonlats.return_value = (1, 2)
        self.source_geo_def = mock.MagicMock(get_lonlats=get_lonlats)
        self.target_geo_def = mock.MagicMock(get_lonlats=get_lonlats)
        self.bucket = BucketSum(self.source_geo_def, self.target_geo_def)

    def _compute_mocked_bucket_sum(self, data, return_data=None, **kwargs):
        """Compute the mocked bucket sum."""
        self.bucket.resampler = mock.MagicMock()
        if return_data is not None:
            self.bucket.resampler.get_sum.return_value = return_data
        else:
            self.bucket.resampler.get_sum.return_value = data
        res = self.bucket.compute(data, **kwargs)
        return res

    def test_compute(self):
        """Test sum bucket resampler computation."""
        # 1D data
        data = da.ones((5,))
        res = self._compute_mocked_bucket_sum(data)
        assert res.shape == (1, 5)
        # 2D data
        data = da.ones((5, 5))
        res = self._compute_mocked_bucket_sum(data)
        assert res.shape == (1, 5, 5)
        # 3D data
        data = da.ones((3, 5, 5))
        res = self._compute_mocked_bucket_sum(data, return_data=data[0, :, :])
        assert res.shape == (3, 5, 5)

    def test_compute_and_use_skipna_handling(self):
        """Test bucket resampler computation and use skipna handling."""
        data = da.ones((5,))

        self._compute_mocked_bucket_sum(data, skipna=False)
        self.bucket.resampler.get_sum.assert_called_once_with(
            data,
            skipna=False)

        self._compute_mocked_bucket_sum(data)
        self.bucket.resampler.get_sum.assert_called_once_with(
            data,
            skipna=True)


class TestBucketCount(unittest.TestCase):
    """Test the count bucket resampler."""

    def setUp(self):
        """Create fake area definitions and resampler to be tested."""
        from satpy.resample.bucket import BucketCount
        get_lonlats = mock.MagicMock()
        get_lonlats.return_value = (1, 2)
        self.source_geo_def = mock.MagicMock(get_lonlats=get_lonlats)
        self.target_geo_def = mock.MagicMock(get_lonlats=get_lonlats)
        self.bucket = BucketCount(self.source_geo_def, self.target_geo_def)

    def _compute_mocked_bucket_count(self, data, return_data=None, **kwargs):
        """Compute the mocked bucket count."""
        self.bucket.resampler = mock.MagicMock()
        if return_data is not None:
            self.bucket.resampler.get_count.return_value = return_data
        else:
            self.bucket.resampler.get_count.return_value = data
        res = self.bucket.compute(data, **kwargs)
        return res

    def test_compute(self):
        """Test count bucket resampler computation."""
        # 1D data
        data = da.ones((5,))
        res = self._compute_mocked_bucket_count(data)
        self.bucket.resampler.get_count.assert_called_once_with()
        assert res.shape == (1, 5)
        # 2D data
        data = da.ones((5, 5))
        res = self._compute_mocked_bucket_count(data)
        self.bucket.resampler.get_count.assert_called_once_with()
        assert res.shape == (1, 5, 5)
        # 3D data
        data = da.ones((3, 5, 5))
        res = self._compute_mocked_bucket_count(data, return_data=data[0, :, :])
        assert res.shape == (3, 5, 5)


class TestBucketFraction(unittest.TestCase):
    """Test the fraction bucket resampler."""

    def setUp(self):
        """Create fake area definitions and resampler to be tested."""
        from satpy.resample.bucket import BucketFraction
        get_lonlats = mock.MagicMock()
        get_lonlats.return_value = (1, 2)
        get_proj_vectors = mock.MagicMock()
        get_proj_vectors.return_value = ([1, 2, 3, 4, 5],  [1, 2, 3, 4, 5])
        self.source_geo_def = mock.MagicMock(get_lonlats=get_lonlats)
        self.target_geo_def = mock.MagicMock(get_lonlats=get_lonlats, crs=None, get_proj_vectors=get_proj_vectors)
        self.bucket = BucketFraction(self.source_geo_def, self.target_geo_def)

    def test_compute(self):
        """Test fraction bucket resampler computation."""
        self.bucket.resampler = mock.MagicMock()
        data = da.ones((3, 3))

        # No kwargs given
        _ = self.bucket.compute(data)
        self.bucket.resampler.get_fractions.assert_called_with(
            data,
            categories=None,
            fill_value=np.nan)
        # Custom kwargs
        _ = self.bucket.compute(data, categories=[1, 2], fill_value=0)
        self.bucket.resampler.get_fractions.assert_called_with(
            data,
            categories=[1, 2],
            fill_value=0)

        # Too many dimensions
        data = da.ones((3, 5, 5))
        with pytest.raises(ValueError, match="BucketFraction not implemented for 3D datasets"):
            _ = self.bucket.compute(data)

    @mock.patch("pyresample.bucket.BucketResampler")
    def test_resample(self, pyresample_bucket):
        """Test fraction bucket resamplers resample method."""
        self.bucket.resampler = mock.MagicMock()
        self.bucket.precompute = mock.MagicMock()
        self.bucket.compute = mock.MagicMock()

        # Fractions return a dict
        data = xr.DataArray(da.ones((1, 5, 5)), dims=("bands", "y", "x"))
        arr = da.ones((5, 5))
        self.bucket.compute.return_value = {0: arr, 1: arr, 2: arr}
        res = self.bucket.resample(data)
        assert "categories" in res.coords
        assert "categories" in res.dims
        assert np.all(res.coords["categories"] == np.array([0, 1, 2]))


@pytest.mark.parametrize("name",
                         ["KDTreeResampler",
                          "BilinearResampler",
                          "NativeResampler",
                          "BucketResamplerBase",
                          "BucketAvg",
                          "BucketSum",
                          "BucketCount",
                          "BucketFraction",
                          "resample",
                          "prepare_resampler",
                          "resample_dataset",
                          "get_area_file",
                          "get_area_def",
                          "add_xy_coords",
                          "add_crs_xy_coords",
                          ]
                         )
def test_moved_import_warns(name):
    """Test that imports done directly from satpy.resample sub-package issue a warning."""
    import satpy.resample
    with pytest.warns(UserWarning, match=".*has been moved.*"):
        _ = getattr(satpy.resample, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Writer for GeoTIFF images with tags for the NinJo visualization tool.

Starting with NinJo 7, NinJo is able to read standard GeoTIFF images,
with required metadata encoded as a set of XML tags in the GDALMetadata
TIFF tag.  Each of the XML tags must be prepended with ``'NINJO_'``.
For NinJo delivery, these GeoTIFF files supersede the old NinJoTIFF
format.  The :class:`NinJoGeoTIFFWriter` therefore supersedes the old
Satpy NinJoTIFF writer and the pyninjotiff package.

The reference documentation for valid NinJo tags and their meaning is
contained in `NinJoPedia`_.  Since this page is not in the public web,
there is a (possibly outdated) `mirror`_.

.. _NinJoPedia: https://ninjopedia.com/tiki-index.php?page=adm_SatelliteServer_SatelliteImportFormats_en
.. _mirror: https://www.ssec.wisc.edu/~davidh/polar2grid/misc/NinJo_Satellite_Import_Formats.html

There are some user-facing differences between the old NinJoTIFF writer
and the new NinJoGeoTIFF writer.  Most notably, keyword arguments that
correspond to tags directly passed by the user are now identical,
including case, to how they will be written to the GDALMetaData and
interpreted by NinJo.  That means some keyword arguments have changed,
such as summarised in this table:

.. list-table:: Migrating to NinJoGeoTIFF, keyword arguments for the writer
   :header-rows: 1

   * - ninjotiff (old)
     - ninjogeotiff (new)
     - Notes
   * - ``chan_id``
     - ``ChannelID``
     - mandatory
   * - ``data_cat``
     - ``DataType``
     - mandatory
   * - ``physic_unit``
     - ``PhysicUnit``
     - mandatory
   * - ``physic_val``
     - ``PhysicValue``
     - mandatory
   * - ``sat_id``
     - ``SatelliteNameID``
     - mandatory
   * - ``data_source``
     - ``DataSource``
     - optional

Moreover, two keyword arguments are no longer supported because
their functionality has become redundant.  This applies to
``ch_min_measurement_unit`` and ``ch_max_measurement_unit``.
Instead, pass those values in source units to the
:func:`~satpy.enhancements.contrast.stretch` enhancement with the ``min_stretch``
and ``max_stretch`` arguments.

For images where the pixel value corresponds directly to a physical value,
NinJo has a functionality to read the corresponding quantity (example:
brightness temperature or reflectance).  To make this possible, the writer
adds the tags ``Gradient`` and ``AxisIntercept``.  Those tags are added if
and only if the image has mode ``L``, ``P``, or ``LA`` and ``PhysicUnit`` is not set
to ``"N/A"``.  In other words, to suppress those tags for images with mode
``L`` or ``LA`` (for example, for the composite ``vis_with_ir``, where the
physical interpretation of individual pixels is lost), one should set
``PhysicUnit`` to ``"N/A"``, ``"n/a"``, ``"1"``, or ``""`` (empty string).
If the image has mode ``P``, ``Gradient`` is set to ``1.0`` and ``AxisIntercept``
to ``0.0`` (as expected by NinJo).
"""

import copy
import datetime
import logging

import numpy as np

from .geotiff import GeoTIFFWriter

logger = logging.getLogger(__name__)


class NinJoGeoTIFFWriter(GeoTIFFWriter):
    """Writer for GeoTIFFs with NinJo tags.

    This writer is experimental.  API may be subject to change.

    For information, see module docstring and documentation for
    :meth:`~NinJoGeoTIFFWriter.save_image`.
    """

    scale_offset_tag_names = ("ninjo_Gradient", "ninjo_AxisIntercept")

    def save_image(  # noqa: D417
            self, image, filename=None, fill_value=None,
            compute=True, keep_palette=False, cmap=None, overviews=None,
            overviews_minsize=256, overviews_resampling=None,
            tags=None, config_files=None,
            *, ChannelID, DataType, PhysicUnit, PhysicValue,
            SatelliteNameID, **kwargs):
        """Save image along with NinJo tags.

        Save image along with NinJo tags.  Interface as for GeoTIFF,
        except NinJo expects some additional tags.  Those tags will be
        prepended with ``ninjo_`` and added as GDALMetaData.

        Writing such images requires trollimage 1.16 or newer.

        Importing such images with NinJo requires NinJo 7 or newer.

        Args:
            image (:class:`~trollimage.xrimage.XRImage`):
                Image to save.
            filename (str): Where to save the file.
            fill_value (int): Which pixel value is fill value?
            compute (bool): To compute or not to compute, that is the question.
            keep_palette (bool):
                As for parent GeoTIFF :meth:`~satpy.writers.geotiff.GeoTIFFWriter.save_image`.
            cmap (:class:`trollimage.colormap.Colormap`):
                As for parent :meth:`~satpy.writers.geotiff.GeoTIFFWriter.save_image`.
            overviews (list):
                As for :meth:`~satpy.writers.geotiff.GeoTIFFWriter.save_image`.
            overviews_minsize (int):
                As for :meth:`~satpy.writers.geotiff.GeoTIFFWriter.save_image`.
            overviews_resampling (str):
                As for :meth:`~satpy.writers.geotiff.GeoTIFFWriter.save_image`.
            tags (dict): Extra (not NinJo) tags to add to GDAL MetaData
            config_files (Any): Not directly used by this writer, supported
                for compatibility with other writers.

        Remaining keyword arguments are either passed as GDAL options,
        if contained in ``self.GDAL_OPTIONS``, or they are passed
        to :class:`NinJoTagGenerator`, which will include them as
        NinJo tags in GDALMetadata.  Supported tags are defined in
        ``NinJoTagGenerator.optional_tags``.  The meaning of those (and
        other) tags are defined in the NinJo documentation (see module
        documentation for a link to NinJoPedia).  The following tags
        are mandatory and must be provided as keyword arguments:

            ChannelID (int)
                NinJo Channel ID
            DataType (int)
                NinJo Data Type
            SatelliteNameID (int)
                NinJo Satellite ID
            PhysicUnit (str)
                NinJo label for unit (example: "C").  If PhysicValue is set to
                "Temperature", PhysicUnit is set to "C", but data attributes
                incidate the data have unit "K", then the writer will adapt the
                header ``ninjo_AxisIntercept`` such that data are interpreted
                in units of "C".  If PhysicUnit is set to "N/A", no
                AxisIntercept and Gradient tags will be written.
            PhysicValue (str)
                NinJo label for quantity (example: "temperature")

        """
        dataset = image.data

        # filename not passed on to writer by Scene.save_dataset, but I need
        # it!
        filename = filename or self.get_filename(**dataset.attrs)

        gdal_opts = {}
        ntg_opts = {}
        for (k, v) in kwargs.items():
            if k in self.GDAL_OPTIONS:
                gdal_opts[k] = v
            else:
                ntg_opts[k] = v

        ntg = NinJoTagGenerator(
            image,
            fill_value=fill_value,
            filename=filename,
            ChannelID=ChannelID,
            DataType=DataType,
            PhysicUnit=PhysicUnit,
            PhysicValue=PhysicValue,
            SatelliteNameID=SatelliteNameID,
            **ntg_opts)
        ninjo_tags = {f"ninjo_{k:s}": v for (k, v) in ntg.get_all_tags().items()}
        image = self._fix_units(image, PhysicValue, PhysicUnit)

        return super().save_image(
            image,
            filename=filename,
            fill_value=fill_value,
            compute=compute,
            keep_palette=keep_palette,
            cmap=cmap,
            overviews=overviews,
            overviews_minsize=overviews_minsize,
            overviews_resampling=overviews_resampling,
            tags={**(tags or {}), **ninjo_tags},
            scale_offset_tags=self._get_scale_offset_tags(image, PhysicUnit),
            **gdal_opts)

    def _get_scale_offset_tags(self, image, unit):
        """Get scale offset tags (tuple or dict)."""
        if self._check_include_scale_offset(image, unit):
            # image.mode cannot be trusted https://github.com/pytroll/satpy/issues/2300
            try:
                mod = image.data.attrs["mode"]
            except KeyError:
                mod = image.mode
            if mod == "P":
                return dict(zip(self.scale_offset_tag_names, (1, 0)))
            return self.scale_offset_tag_names
        return None  # explicit is better than implicit

    def _fix_units(self, image, quantity, unit):
        """Adapt units between °C and K.

        This will return a new XRImage, to make sure the old data and
        enhancement history aren't touched.
        """
        data_units = image.data.attrs.get("units")
        if (quantity.lower() == "temperature" and
                unit == "C" and
                data_units == "K"):
            logger.debug("Adding offset for K → °C conversion")
            new_attrs = copy.deepcopy(image.data.attrs)
            im2 = type(image)(image.data.copy())
            im2.data.attrs = new_attrs
            # this scale/offset has to be applied before anything else
            im2.data.attrs["enhancement_history"].insert(0, {"scale": 1, "offset": 273.15})
            return im2
        if unit != data_units and unit.lower() != "n/a":
            logger.warning(
                    f"Writing {unit!s} to ninjogeotiff headers, but "
                    f"data attributes have unit {data_units!s}. "
                    "No conversion applied.")

        return image

    def _check_include_scale_offset(self, image, unit):
        """Check if scale-offset tags should be included."""
        if image.mode[0] in "LP" and unit.lower() not in ("n/a", "1", ""):
            return True
        return False


class NinJoTagGenerator:
    """Class to collect NinJo tags.

    This class is used by :class:`NinJoGeoTIFFWriter` to collect NinJo tags.
    Most end-users will not need to create instances of this class directly.

    Tags are gathered from three sources:

    - Fixed tags, contained in the attribute ``fixed_tags``.  The value of
      those tags is hardcoded and never changes.
    - Tags passed by the user, contained in the attribute ``passed_tags``.
      Those tags must be passed by the user as arguments to the writer, which
      will pass them on when instantiating this class.
    - Tags calculated from data and metadata.  Those tags are defined in the
      attribute ``dynamic_tags``.  They are either calculated from image data,
      from image metadata, or from arguments passed by the user to the writer.

    Some tags are mandatory (defined in ``mandatory_tags``).  All tags that are
    not mandatory are optional.  By default, optional tags are generated if and
    only if the required information is available.
    """

    # tags that never change
    fixed_tags = {
        "Magic": "NINJO",
        "HeaderVersion": 2,
        "XMinimum": 1,
        "YMinimum": 1}

    # tags that must be passed directly by the user
    passed_tags = {"ChannelID", "DataType", "PhysicUnit",
                   "SatelliteNameID", "PhysicValue"}

    # tags that can be calculated dynamically from (meta)data
    dynamic_tags = {
        "CentralMeridian": "central_meridian",
        "ColorDepth": "color_depth",
        "CreationDateID": "creation_date_id",
        "DateID": "date_id",
        "EarthRadiusLarge": "earth_radius_large",
        "EarthRadiusSmall": "earth_radius_small",
        "FileName": "filename",
        "MaxGrayValue": "max_gray_value",
        "MinGrayValue": "min_gray_value",
        "Projection": "projection",
        "ReferenceLatitude1": "ref_lat_1",
        "TransparentPixel": "transparent_pixel",
        "XMaximum": "xmaximum",
        "YMaximum": "ymaximum"
        }

    # mandatory tags according to documentation
    mandatory_tags = {"SatelliteNameID", "DateID", "CreationDateID",
                      "ChannelID", "HeaderVersion", "DataType",
                      "SatelliteNumber", "ColorDepth", "XMinimum", "XMaximum",
                      "YMinimum", "YMaximum", "Projection", "PhysicValue",
                      "PhysicUnit", "MinGrayValue", "MaxGrayValue", "Gradient",
                      "AxisIntercept", "TransparentPixel"}

    # optional tags are added on best effort or if passed by user
    optional_tags = {"DataSource", "MeridianWest", "MeridianEast",
                     "EarthRadiusLarge", "EarthRadiusSmall", "GeodeticDate",
                     "ReferenceLatitude1", "ReferenceLatitude2",
                     "CentralMeridian", "ColorTable", "Description",
                     "OverflightDirection", "GeoLatitude", "GeoLongitude",
                     "Altitude", "AOSAzimuth", "LOSAzimuth", "MaxElevation",
                     "OverFlightTime", "IsBlackLinesCorrection",
                     "IsAtmosphereCorrected", "IsCalibrated", "IsNormalized",
                     "OriginalHeader", "IsValueTableAvailable",
                     "ValueTableFloatField"}

    # tags that are added later in other ways
    postponed_tags = {"AxisIntercept", "Gradient"}

    def __init__(self, image, fill_value, filename, **kwargs):
        """Initialise tag generator.

        Args:
            image (:class:`trollimage.xrimage.XRImage`): XRImage for which
                NinJo tags should be calculated.
            fill_value (int): Fill value corresponding to image.
            filename (str): Filename to be written.
            **kwargs: Any additional tags to be included as-is.
        """
        self.image = image
        self.dataset = image.data
        self.fill_value = fill_value
        self.filename = filename
        self.args = kwargs
        self.tag_names = (self.fixed_tags.keys() |
                          self.passed_tags |
                          self.dynamic_tags.keys() |
                          (self.args.keys() & self.optional_tags))
        if self.args.keys() - self.tag_names:
            raise ValueError("The following tags were not recognised: " +
                             " ".join(self.args.keys() - self.tag_names))

    def get_all_tags(self):
        """Get a dictionary with all tags for NinJo."""
        tags = {}
        for tag in self.tag_names:
            try:
                tags[tag] = self.get_tag(tag)
            except (AttributeError, KeyError) as e:
                if tag in self.mandatory_tags:
                    raise
                logger.debug(
                    f"Unable to obtain value for optional NinJo tag {tag:s}. "
                    f"This is probably expected.  The reason is: {e.args[0]}")
        return tags

    def get_tag(self, tag):
        """Get value for NinJo tag."""
        if tag in self.fixed_tags:
            return self.fixed_tags[tag]
        if tag in self.passed_tags:
            return self.args[tag]
        if tag in self.dynamic_tags:
            return getattr(self, f"get_{self.dynamic_tags[tag]:s}")()
        if tag in self.optional_tags and tag in self.args:
            return self.args[tag]
        if tag in self.postponed_tags:
            raise ValueError(f"Tag {tag!s} is added later by the GeoTIFF writer.")
        if tag in self.optional_tags:
            raise ValueError(
                f"Optional tag {tag!s} must be supplied by user if user wants to "
                "request the value, but wasn't.")
        raise ValueError(f"Unknown tag: {tag!s}")

    def get_central_meridian(self):
        """Calculate central meridian."""
        pams = self.dataset.attrs["area"].crs.coordinate_operation.params
        lon_0 = {p.name: p.value for p in pams}["Longitude of natural origin"]
        return lon_0

    def get_color_depth(self):
        """Return the color depth."""
        if self.image.mode in ("L", "P"):
            return 8
        if self.image.mode in ("LA", "PA"):
            return 16
        if self.image.mode == "RGB":
            return 24
        if self.image.mode == "RGBA":
            return 32
        raise ValueError(
                f"Unsupported image mode: {self.image.mode:s}")

    # Set unix epoch here explicitly, because datetime.timestamp() is
    # apparently not supported on Windows.
    _epoch = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)

    def get_creation_date_id(self):
        """Calculate the creation date ID.

        That's seconds since UNIX Epoch for the time the image is created.
        """
        delta = datetime.datetime.now(tz=datetime.timezone.utc) - self._epoch
        return int(delta.total_seconds())

    def get_date_id(self):
        """Calculate the date ID.

        That's seconds since UNIX Epoch for the time corresponding to the
        satellite image start of measurement time.
        """
        tm = self.dataset.attrs["start_time"]
        delta = tm.replace(tzinfo=datetime.timezone.utc) - self._epoch
        return int(delta.total_seconds())

    def get_earth_radius_large(self):
        """Return the Earth semi-major axis in metre."""
        return self.dataset.attrs["area"].crs.ellipsoid.semi_major_metre

    def get_earth_radius_small(self):
        """Return the Earth semi-minor axis in metre."""
        return self.dataset.attrs["area"].crs.ellipsoid.semi_minor_metre

    def get_filename(self):
        """Return the filename."""
        return self.filename

    def get_min_gray_value(self):
        """Calculate minimum gray value."""
        return self.image._scale_to_dtype(
            self.dataset.min(keep_attrs=False),
            np.uint8,
            self.fill_value).astype(np.uint8)

    def get_max_gray_value(self):
        """Calculate maximum gray value."""
        return self.image._scale_to_dtype(
            self.dataset.max(keep_attrs=False),
            np.uint8,
            self.fill_value).astype(np.uint8)

    def get_projection(self):
        """Get NinJo projection string.

        From the documentation, valid values are:

        - NPOL/SPOL: polar-sterographic North/South
        - PLAT: „Plate Carrée“, equirectangular projection
        - MERC: Mercator projection

        Derived from AreaDefinition.
        """
        if self.dataset.attrs["area"].crs.coordinate_system.name == "ellipsoidal":
            # For lat/lon coordinates, we say it's PLAT
            return "PLAT"
        name = self.dataset.attrs["area"].crs.coordinate_operation.method_name
        if "Equidistant Cylindrical" in name:
            return "PLAT"
        if "Mercator" in name:
            return "MERC"
        if "Stereographic" in name:
            if self.get_ref_lat_1() >= 0:
                return "NPOL"
            return "SPOL"
        raise ValueError(
                "Unknown mapping from area "
                f"{self.dataset.attrs['area'].description!r} with CRS coordinate "
                f"operation name {name:s} to NinJo projection.  NinJo understands only "
                "equidistant cylindrical, mercator, or stereographic projections.")

    def get_ref_lat_1(self):
        """Get reference latitude one.

        Derived from area definition.
        """
        pams = {p.name: p.value for p in self.dataset.attrs["area"].crs.coordinate_operation.params}
        for label in ["Latitude of standard parallel",
                      "Latitude of natural origin",
                      "Latitude of 1st standard parallel"]:
            if label in pams:
                return pams[label]
        raise ValueError(
                "Could not find reference latitude for area "
                f"{self.dataset.attrs['area'].description}")

    def get_transparent_pixel(self):
        """Get the transparent pixel value, also known as the fill value.

        When the no fill value is defined (value `None`), such as for RGBA or
        LA images, returns -1, in accordance with the file format
        specification.
        """
        if self.fill_value is None:
            return -1
        return self.fill_value

    def get_xmaximum(self):
        """Get the maximum value of x, i.e. the meridional extent of the image in pixels."""
        return self.dataset.sizes["x"]

    def get_ymaximum(self):
        """Get the maximum value of y, i.e. the zonal extent of the image in pixels."""
        return self.dataset.sizes["y"]

    def get_meridian_east(self):
        """Get the easternmost longitude of the area.

        Currently not implemented.  In pyninjotiff it was implemented but the
        answer was incorrect.
        """
        raise NotImplementedError()

    def get_meridian_west(self):
        """Get the westernmost longitude of the area.

        Currently not implemented.  In pyninjotiff it was implemented but the
        answer was incorrect.
        """
        raise NotImplementedError()

    def get_ref_lat_2(self):
        """Get reference latitude two.

        This is not implemented and never was correctly implemented in
        pyninjotiff either.  It doesn't appear to be used by NinJo.
        """
        raise NotImplementedError("Second reference latitude not implemented.")



# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""CF processing of pyresample area information."""
import logging

import xarray as xr
from packaging.version import Version
from pyresample.geometry import AreaDefinition, SwathDefinition

logger = logging.getLogger(__name__)


def _add_lonlat_coords(data_arr: xr.DataArray) -> xr.DataArray:
    """Add 'longitude' and 'latitude' coordinates to DataArray."""
    data_arr = data_arr.copy()
    area = data_arr.attrs["area"]
    ignore_dims = {dim: 0 for dim in data_arr.dims if dim not in ["x", "y"]}
    chunks = getattr(data_arr.isel(**ignore_dims), "chunks", None)
    lons, lats = area.get_lonlats(chunks=chunks)
    data_arr["longitude"] = xr.DataArray(lons, dims=["y", "x"],
                                         attrs={"name": "longitude",
                                                 "standard_name": "longitude",
                                                 "units": "degrees_east"},
                                         name="longitude")
    data_arr["latitude"] = xr.DataArray(lats, dims=["y", "x"],
                                        attrs={"name": "latitude",
                                                "standard_name": "latitude",
                                                "units": "degrees_north"},
                                        name="latitude")
    return data_arr


def _create_grid_mapping(area):
    """Create the grid mapping instance for `area`."""
    import pyproj

    if Version(pyproj.__version__) < Version("2.4.1"):
        # technically 2.2, but important bug fixes in 2.4.1
        raise ImportError("'cf' writer requires pyproj 2.4.1 or greater")
    # let pyproj do the heavily lifting (pyproj 2.0+ required)
    grid_mapping = area.crs.to_cf()
    return area.area_id, grid_mapping


def _add_grid_mapping(data_arr: xr.DataArray) -> tuple[xr.DataArray, xr.DataArray]:
    """Convert an area to at CF grid mapping."""
    data_arr = data_arr.copy()
    area = data_arr.attrs["area"]
    gmapping_var_name, attrs = _create_grid_mapping(area)
    data_arr.attrs["grid_mapping"] = gmapping_var_name
    return data_arr, xr.DataArray(0, attrs=attrs, name=gmapping_var_name)


def area2cf(data_arr: xr.DataArray, include_lonlats: bool = False, got_lonlats: bool = False) -> list[xr.DataArray]:
    """Convert an area to at CF grid mapping or lon and lats."""
    res = []
    include_lonlats = include_lonlats or isinstance(data_arr.attrs["area"], SwathDefinition)
    is_area_def = isinstance(data_arr.attrs["area"], AreaDefinition)
    if not got_lonlats and include_lonlats:
        data_arr = _add_lonlat_coords(data_arr)
    if is_area_def:
        data_arr, gmapping = _add_grid_mapping(data_arr)
        res.append(gmapping)
    res.append(data_arr)
    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittesting the native msg reader."""

import datetime as dt

import numpy as np
import pytest
import xarray as xr
from pytest_lazy_fixtures.lazy_fixture import lf

import satpy.readers.core.seviri as sev

COUNTS_INPUT = xr.DataArray(
    np.array([[377.,  377.,  377.,  376.,  375.],
              [376.,  375.,  376.,  374.,  374.],
              [374.,  373.,  373.,  374.,  374.],
              [347.,  345.,  345.,  348.,  347.],
              [306.,  306.,  307.,  307.,  308.]], dtype=np.float32)
)

RADIANCES_OUTPUT = xr.DataArray(
    np.array([[66.84162903,  66.84162903,  66.84162903,  66.63659668,
               66.4315567],
              [66.63659668,  66.4315567,  66.63659668,  66.22652435,
               66.22652435],
              [66.22652435,  66.02148438,  66.02148438,  66.22652435,
               66.22652435],
              [60.69055939,  60.28048706,  60.28048706,  60.89559937,
               60.69055939],
              [52.28409576,  52.28409576,  52.48912811,  52.48912811,
               52.69416809]], dtype=np.float32)
)

GAIN = 0.20503567620766011
OFFSET = -10.456819486590666
CHANNEL_NAME = "IR_108"
PLATFORM_ID = 323  # Met-10

TBS_OUTPUT1 = xr.DataArray(
    np.array([[269.29684448,  269.29684448,  269.29684448,  269.13296509,
               268.96871948],
              [269.13296509,  268.96871948,  269.13296509,  268.80422974,
               268.80422974],
              [268.80422974,  268.63937378,  268.63937378,  268.80422974,
               268.80422974],
              [264.23751831,  263.88912964,  263.88912964,  264.41116333,
               264.23751831],
              [256.77682495,  256.77682495,  256.96743774,  256.96743774,
               257.15756226]], dtype=np.float32)
)

TBS_OUTPUT2 = xr.DataArray(
    np.array([[268.94519043,  268.94519043,  268.94519043,  268.77984619,
               268.61422729],
              [268.77984619,  268.61422729,  268.77984619,  268.44830322,
               268.44830322],
              [268.44830322,  268.28204346,  268.28204346,  268.44830322,
               268.44830322],
              [263.84396362,  263.49285889,  263.49285889,  264.01898193,
               263.84396362],
              [256.32858276,  256.32858276,  256.52044678,  256.52044678,
               256.71188354]], dtype=np.float32)
)

VIS008_SOLAR_IRRADIANCE = 73.1807

VIS008_RADIANCE = xr.DataArray(
    np.array([[0.62234485,  0.59405649,  0.59405649,  0.59405649,  0.59405649],
             [0.59405649,  0.62234485,  0.62234485,  0.59405649,  0.62234485],
             [0.76378691,  0.79207528,  0.79207528,  0.76378691,  0.79207528],
             [3.30974245,  3.33803129,  3.33803129,  3.25316572,  3.47947311],
             [7.52471399,  7.83588648,  8.2602129,  8.57138538,  8.99571133]],
             dtype=np.float32)
)

VIS008_REFLECTANCE = xr.DataArray(
    np.array([[2.739768, 2.615233, 2.615233, 2.615233, 2.615233],
              [2.615233, 2.739768, 2.739768, 2.615233, 2.739768],
              [3.362442, 3.486977, 3.486977, 3.362442, 3.486977],
              [14.570578, 14.695117, 14.695117, 14.321507, 15.317789],
              [33.126278, 34.49616, 36.364185, 37.73407, 39.60209]],
             dtype=np.float32)
)


class TestSEVIRICalibrationAlgorithm:
    """Unit Tests for SEVIRI calibration algorithm."""

    @pytest.fixture
    def algo(self):
        """Set up the SEVIRI Calibration algorithm for testing."""
        return sev.SEVIRICalibrationAlgorithm(
            platform_id=PLATFORM_ID,
            scan_time=dt.datetime(2020, 8, 15, 13, 0, 40)
        )

    def test_convert_to_radiance(self, algo):
        """Test the conversion from counts to radiances."""
        result = algo.convert_to_radiance(COUNTS_INPUT, GAIN, OFFSET)
        xr.testing.assert_allclose(result, RADIANCES_OUTPUT)
        assert result.dtype == np.float32

    def test_ir_calibrate_spectral_radiance(self, algo):
        """Test conversion from spectral radiance to brightness temperature."""
        result = algo.ir_calibrate(RADIANCES_OUTPUT, CHANNEL_NAME,
                                   sev.IRCalibrationType.spectral_radiance)
        xr.testing.assert_allclose(result, TBS_OUTPUT1, rtol=1E-5)
        assert result.dtype == np.float32

    def test_ir_calibrate_effective_radiance(self, algo):
        """Test conversion from effective radiance to brightness temperature."""
        result = algo.ir_calibrate(RADIANCES_OUTPUT, CHANNEL_NAME,
                                   sev.IRCalibrationType.effective_radiance)
        xr.testing.assert_allclose(result, TBS_OUTPUT2, rtol=1E-5)

    def test_ir_calibrate_missing_coefs(self, algo):
        """Test IR calibration with missing coefficients."""
        with pytest.raises(ValueError, match="No calibration coefficients"):
            algo.ir_calibrate(RADIANCES_OUTPUT, CHANNEL_NAME,
                              sev.IRCalibrationType.not_processed)

    def test_ir_calibrate_bad_calib_type(self, algo):
        """Test IR calibration with bad calibration type."""
        with pytest.raises(NotImplementedError):
            algo.ir_calibrate(RADIANCES_OUTPUT, CHANNEL_NAME, cal_type=-1)

    def test_vis_calibrate(self, algo):
        """Test conversion from radiance to reflectance."""
        result = algo.vis_calibrate(VIS008_RADIANCE, VIS008_SOLAR_IRRADIANCE)
        xr.testing.assert_allclose(result, VIS008_REFLECTANCE)
        assert result.sun_earth_distance_correction_applied
        assert result.dtype == np.float32


class TestSeviriCalibrationHandler:
    """Unit tests for SEVIRI calibration handler."""

    def test_init(self):
        """Test initialization of the calibration handler."""
        with pytest.raises(ValueError, match="Invalid calibration mode: INVALID. Choose one of (.*)"):
            self._get_calibration_handler("IR_108", "INVALID")

    def _get_calibration_handler(self, channel, calib_mode="NOMINAL", ext_coefs=None):
        """Provide a calibration handler."""
        int_coefs = {
                "NOMINAL": {
                    "IR_108": {
                        "gain": 10,
                        "offset": -1
                    },
                    "VIS006": {
                        "gain": 20,
                        "offset": -2
                    },
                },
                "GSICS": {
                    "IR_108": {
                        "gain": 30,
                        "offset": -3
                    },
                },
            }
        calib_params = sev.CalibParams(
            mode=calib_mode,
            internal_coefs=int_coefs,
            external_coefs=ext_coefs,
            radiance_type=1)
        scan_params = sev.ScanParams(
            platform_id=324,
            channel_name=channel,
            scan_time=None
        )
        return sev.SEVIRICalibrationHandler(calib_params, scan_params)

    def test_calibrate_exceptions(self):
        """Test exceptions raised by the calibration handler."""
        calib = self._get_calibration_handler("IR_108")
        with pytest.raises(ValueError, match="Invalid calibration invalid for channel IR_108"):
            calib.calibrate(None, "invalid")

    @pytest.fixture
    def external_coefs(self):
        """Get external coefficients."""
        return {"IR_108": {"gain": 40, "offset": -4}}

    @pytest.fixture
    def coefs_ir108_nominal_exp(self):
        """Get expected IR coefficients in nominal calib mode."""
        return {"coefs": {"gain": 10, "offset": -1}, "mode": "NOMINAL"}

    @pytest.fixture
    def coefs_vis006_exp(self):
        """Get expected VIS coefficients."""
        return {"coefs": {"gain": 20, "offset": -2}, "mode": "NOMINAL"}

    @pytest.fixture
    def coefs_ir108_gsics_exp(self):
        """Get expected IR coefficients in GSICS calib mode."""
        return {"coefs": {"gain": 30, "offset": -3}, "mode": "GSICS"}

    @pytest.fixture
    def coefs_ir108_external_exp(self):
        """Get expected IR coefficients in the presence of external coefficients."""
        return {"coefs": {"gain": 40, "offset": -4}, "mode": "external"}

    @pytest.mark.parametrize(
        ("channel", "calib_mode", "ext_coefs", "expected"),
        [
            ("IR_108", "NOMINAL", None, lf("coefs_ir108_nominal_exp")),
            ("IR_108", "GSICS", None, lf("coefs_ir108_gsics_exp")),
            ("IR_108", "NOMINAL", lf("external_coefs"), lf("coefs_ir108_external_exp")),
            # For VIS006 there's only nominal coefficients in this example
            ("VIS006", "NOMINAL", None, lf("coefs_vis006_exp")),
            ("VIS006", "GSICS", None, lf("coefs_vis006_exp")),
            ("VIS006", "NOMINAL", lf("external_coefs"), lf("coefs_vis006_exp"))
        ]
    )
    def test_get_coefs(self, channel, calib_mode, ext_coefs, expected):
        """Test selection of gain and offset."""
        calib = self._get_calibration_handler(channel, calib_mode, ext_coefs)
        coefs = calib.get_coefs()
        assert coefs == expected


class TestFileHandlerCalibrationBase:
    """Base class for file handler calibration tests."""

    platform_id = 324
    gains_nominal = np.arange(1, 13)
    offsets_nominal = np.arange(-1, -13, -1)
    # No GSICS coefficients for VIS channels -> set to zero
    gains_gsics = [0, 0, 0, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 0]
    offsets_gsics = [0, 0, 0, -0.4, -0.5, -0.6, -0.7, -0.8, -0.9, -1.0, -1.1, 0]
    radiance_types = 2 * np.ones(12)
    scan_time = dt.datetime(2020, 1, 1)
    external_coefs = {
        "VIS006": {"gain": 10, "offset": -10},
        "IR_108": {"gain": 20, "offset": -20},
        "HRV": {"gain": 5, "offset": -5}
    }
    spectral_channel_ids = {"VIS006": 1, "IR_108": 9, "HRV": 12}
    expected = {
        "VIS006": {
            "counts": {
                "NOMINAL": xr.DataArray(
                    [[0, 10],
                     [100, 255]],
                    dims=("y", "x")
                )
            },
            "radiance": {
                "NOMINAL": xr.DataArray(
                    [[np.nan, 9],
                     [99, 254]],
                    dims=("y", "x")
                ),
                "GSICS": xr.DataArray(
                    [[np.nan, 9],
                     [99, 254]],
                    dims=("y", "x")
                ),
                "EXTERNAL": xr.DataArray(
                    [[np.nan, 90],
                     [990, 2540]],
                    dims=("y", "x")
                )
            },
            "reflectance": {
                "NOMINAL": xr.DataArray(
                    [[np.nan, 41.88985],
                     [460.7884, 1182.2247]],
                    dims=("y", "x")
                ),
                "EXTERNAL": xr.DataArray(
                    [[np.nan, 418.89853],
                     [4607.8843, 11822.249]],
                    dims=("y", "x")
                )
            }
        },
        "IR_108": {
            "counts": {
                "NOMINAL": xr.DataArray(
                    [[0, 10],
                     [100, 255]],
                    dims=("y", "x")
                )
            },
            "radiance": {
                "NOMINAL": xr.DataArray(
                    [[np.nan, 81],
                     [891, 2286]],
                    dims=("y", "x")
                ),
                "GSICS": xr.DataArray(
                    [[np.nan, 8.19],
                     [89.19, 228.69]],
                    dims=("y", "x")
                ),
                "EXTERNAL": xr.DataArray(
                    [[np.nan, 180],
                     [1980, 5080]],
                    dims=("y", "x")
                )
            },
            "brightness_temperature": {
                "NOMINAL": xr.DataArray(
                    [[np.nan, 279.82318],
                     [543.2585, 812.77167]],
                    dims=("y", "x")
                ),
                "GSICS": xr.DataArray(
                    [[np.nan, 189.20985],
                     [285.53293, 356.06668]],
                    dims=("y", "x")
                ),
                "EXTERNAL": xr.DataArray(
                    [[np.nan, 335.14236],
                     [758.6249, 1262.7567]],
                    dims=("y", "x")
                ),
            }
        },
        "HRV": {
            "counts": {
                "NOMINAL": xr.DataArray(
                    [[0, 10],
                     [100, 255]],
                    dims=("y", "x")
                )
            },
            "radiance": {
                "NOMINAL": xr.DataArray(
                    [[np.nan, 108],
                     [1188, 3048]],
                    dims=("y", "x")
                ),
                "GSICS": xr.DataArray(
                    [[np.nan, 108],
                     [1188, 3048]],
                    dims=("y", "x")
                ),
                "EXTERNAL": xr.DataArray(
                    [[np.nan, 45],
                     [495, 1270]],
                    dims=("y", "x")
                )
            },
            "reflectance": {
                "NOMINAL": xr.DataArray(
                    [[np.nan, 415.26767],
                     [4567.944, 11719.775]],
                    dims=("y", "x")
                ),
                "EXTERNAL": xr.DataArray(
                    [[np.nan, 173.02817],
                     [1903.31, 4883.2397]],
                    dims=("y", "x")
                )
            }
        }
    }

    @pytest.fixture(name="counts")
    def counts(self):
        """Provide fake image counts."""
        return xr.DataArray(
            [[0, 10],
             [100, 255]],
            dims=("y", "x")
        )

    def _get_expected(
            self, channel, calibration, calib_mode, use_ext_coefs
    ):
        if use_ext_coefs:
            return self.expected[channel][calibration]["EXTERNAL"]
        return self.expected[channel][calibration][calib_mode]



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the writer compute module."""

from __future__ import annotations

import datetime as dt
import os

import dask
import numpy as np
import pytest
import xarray as xr
from dask import array as da

from satpy.writers.core.compute import compute_writer_results


def test_group_results_by_output_file(tmp_path):
    """Test grouping results by output file.

    Add a test for grouping the results from save_datasets(..., compute=False)
    by output file.  This is useful if for some reason we want to treat each
    output file as a seperate computation (that can still be computed together
    later).
    """
    from pyresample import create_area_def

    from satpy.tests.utils import make_fake_scene
    from satpy.writers.core.compute import group_results_by_output_file

    x = 10
    fake_area = create_area_def("sargasso", 4326, resolution=1, width=x, height=x, center=(0, 0))
    fake_scene = make_fake_scene(
        {
            "dragon_top_height": (dat := xr.DataArray(dims=("y", "x"), data=da.arange(x * x).reshape((x, x)))),
            "penguin_bottom_height": dat,
            "kraken_depth": dat,
        },
        daskify=True,
        area=fake_area,
        common_attrs={"start_time": dt.datetime(2022, 11, 16, 13, 27)},
    )
    # NB: even if compute=False, ``save_datasets`` creates (empty) files
    (sources, targets) = fake_scene.save_datasets(
        filename=os.fspath(tmp_path / "test-{name}.tif"),
        writer="ninjogeotiff",
        compress="NONE",
        fill_value=0,
        compute=False,
        ChannelID="x",
        DataType="x",
        PhysicUnit="K",
        PhysicValue="Temperature",
        SatelliteNameID="x",
    )

    grouped = group_results_by_output_file(sources, targets)

    assert len(grouped) == 3
    assert len({x.rfile.path for x in grouped[0][1]}) == 1
    for x in grouped:
        assert len(x[0]) == len(x[1])
    assert sources[:5] == grouped[0][0]
    assert targets[:5] == grouped[0][1]
    assert sources[10:] == grouped[2][0]
    assert targets[10:] == grouped[2][1]


@pytest.fixture
def fake_scene():
    """Create a fake Scene object for testing computing delayed results."""
    from pyresample.geometry import AreaDefinition

    from satpy.scene import Scene

    adef = AreaDefinition(
        "test",
        "test",
        "test",
        "EPSG:4326",
        100,
        200,
        (-180.0, -90.0, 180.0, 90.0),
    )
    ds1 = xr.DataArray(
        da.arange(100 * 200).reshape((100, 200)).rechunk(50),
        dims=("y", "x"),
        attrs={"name": "test", "start_time": dt.datetime(2018, 1, 1, 0, 0, 0), "area": adef},
    )
    scn = Scene()
    scn["test"] = ds1
    return scn


def test_simple_image(tmp_path, fake_scene):
    """Test writing to PNG file."""
    fname = str(tmp_path / "simple_image.png")
    res = fake_scene.save_datasets(
        filename=fname,
        datasets=["test"],
        writer="simple_image",
        compute=False,
    )
    compute_writer_results([res])
    assert os.path.isfile(fname)


def test_geotiff(tmp_path, fake_scene):
    """Test writing to mitiff file."""
    fname = str(tmp_path / "geotiff.tif")
    res = fake_scene.save_datasets(filename=fname, datasets=["test"], writer="geotiff", compute=False)
    compute_writer_results([res])
    assert os.path.isfile(fname)


def test_multiple_geotiff(tmp_path, fake_scene):
    """Test writing to mitiff file."""
    fname1 = str(tmp_path / "geotiff1.tif")
    res1 = fake_scene.save_datasets(filename=fname1, datasets=["test"], writer="geotiff", compute=False)
    fname2 = str(tmp_path / "geotiff2.tif")
    res2 = fake_scene.save_datasets(filename=fname2, datasets=["test"], writer="geotiff", compute=False)
    compute_writer_results([res1, res2])
    assert os.path.isfile(fname1)
    assert os.path.isfile(fname2)


def test_multiple_simple(tmp_path, fake_scene):
    """Test writing to geotiff files."""
    fname1 = str(tmp_path / "simple_image1.png")
    res1 = fake_scene.save_datasets(filename=fname1, datasets=["test"], writer="simple_image", compute=False)
    fname2 = str(tmp_path / "simple_image2.png")
    res2 = fake_scene.save_datasets(filename=fname2, datasets=["test"], writer="simple_image", compute=False)
    compute_writer_results([res1, res2])
    assert os.path.isfile(fname1)
    assert os.path.isfile(fname2)


def test_mixed(tmp_path, fake_scene):
    """Test writing to multiple mixed-type files."""
    fname1 = str(tmp_path / "simple_image3.png")
    res1 = fake_scene.save_datasets(filename=fname1, datasets=["test"], writer="simple_image", compute=False)
    fname2 = str(tmp_path / "geotiff3.tif")
    res2 = fake_scene.save_datasets(filename=fname2, datasets=["test"], writer="geotiff", compute=False)
    res3 = []
    compute_writer_results([res1, res2, res3])
    assert os.path.isfile(fname1)
    assert os.path.isfile(fname2)


def test_source_only():
    """Test writers who only return dask arrays with no targets.

    With newer versions of dask it is recommended to not pass Arrays to
    Delayed functions as the tasks aren't properly/completely optimized.
    The alternative is to reduce the Array into a single array-like operation
    with map_blocks, blockwise, or some other reduction function.
    """
    compute_count = 0

    def reduced_map_blocks(_: np.ndarray) -> str:
        nonlocal compute_count
        compute_count += 1
        return "abc"

    arr1 = da.zeros((5, 5))
    arr2 = da.ones((5, 5))
    res1 = da.map_blocks(
        reduced_map_blocks,
        arr1.rechunk(arr1.shape),
        dtype=str,
        meta=np.ndarray((), dtype=str),
        chunks=(1, 1),
    )
    res2 = da.map_blocks(
        reduced_map_blocks,
        arr2.rechunk(arr2.shape),
        dtype=str,
        meta=np.ndarray((), dtype=str),
        chunks=(1, 1),
    )

    compute_writer_results([[res1], [res2]])
    assert compute_count == 2


@pytest.mark.parametrize(
    "results",
    [
        [],
        [[]],
    ],
)
def test_empty(results):
    """Test empty result list."""
    from satpy.tests.utils import assert_maximum_dask_computes

    # almost impossible for dask to compute anything, but let's make sure
    with assert_maximum_dask_computes(0):
        res = compute_writer_results(results)
        assert res == []


class _Writable:
    def __setitem__(self, window_slice, data): ...


TEST_TARGET = _Writable()
TEST_SRC = da.zeros((5, 5), chunks=2)


@pytest.mark.parametrize(
    "results",
    [
        (TEST_SRC, TEST_TARGET),
        [(TEST_SRC, TEST_TARGET)],
        [dask.delayed(TEST_SRC)],
    ],
)
def test_legacy_return_values(results):
    """Test old ways that writers can return things produces a warning."""
    # in Satpy 1.0, change to `pytest.raises`
    with pytest.warns(UserWarning, match="Unexpected result from Satpy writer"):
        compute_writer_results(results)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""SCMI NetCDF4 Reader.

SCMI files are typically used for data for the ABI instrument onboard the
GOES-16/17 satellites. It is the primary format used for providing ABI data
to the AWIPS visualization clients used by the US National Weather Service
forecasters. The python code for this reader may be reused by other readers
as NetCDF schemes/metadata change for different products. The initial reader
using this code is the "scmi_abi" reader (see `abi_l1b_scmi.yaml` for more
information).

There are two forms of these files that this reader supports:

1. Official SCMI format: NetCDF4 files where the main data variable is stored
    in a variable called "Sectorized_CMI". This variable name can be
    configured in the YAML configuration file.
2. Satpy/Polar2Grid SCMI format: NetCDF4 files based on the official SCMI
    format created for the Polar2Grid project. This format was migrated to
    Satpy as part of Polar2Grid's adoption of Satpy for the majority of its
    features. This format is what is produced by Satpy's `scmi` writer.
    This format can be identified by a single variable named "data" and a
    global attribute named ``"awips_id"`` that is set to a string starting with
    ``"AWIPS_"``.

"""

import datetime as dt
import logging
import os

import numpy as np
import xarray as xr
from pyresample import geometry

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()
# NetCDF doesn't support multi-threaded reading, trick it by opening
# as one whole chunk then split it up before we do any calculations
LOAD_CHUNK_SIZE = int(os.getenv("PYTROLL_LOAD_CHUNK_SIZE", -1))
logger = logging.getLogger(__name__)


class SCMIFileHandler(BaseFileHandler):
    """Handle a single SCMI NetCDF4 file."""

    def __init__(self, filename, filename_info, filetype_info):
        """Set up the SCMI file handler."""
        super(SCMIFileHandler, self).__init__(filename, filename_info,
                                              filetype_info)
        # xarray's default netcdf4 engine
        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=False,
                                  chunks={"x": LOAD_CHUNK_SIZE, "y": LOAD_CHUNK_SIZE})
        self.platform_name = self.nc.attrs["satellite_id"]
        self.sensor = self._get_sensor()
        self.nlines = self.nc.dims["y"]
        self.ncols = self.nc.dims["x"]
        self.coords = {}

    def _get_sensor(self):
        """Determine the sensor for this file."""
        # sometimes Himawari-8 (or 9) data is stored in SCMI format
        is_h8 = "H8" in self.platform_name
        is_h9 = "H9" in self.platform_name
        is_ahi = is_h8 or is_h9
        return "ahi" if is_ahi else "abi"

    @property
    def sensor_names(self):
        """Get the sensor names."""
        return [self.sensor]

    def __getitem__(self, item):
        """Wrap around `self.nc[item]`.

        Some datasets use a 32-bit float scaling factor like the 'x' and 'y'
        variables which causes inaccurate unscaled data values. This method
        forces the scale factor to a 64-bit float first.

        """
        data = self.nc[item]
        attrs = data.attrs
        factor = data.attrs.get("scale_factor")
        offset = data.attrs.get("add_offset")
        fill = data.attrs.get("_FillValue")
        if fill is not None:
            data = data.where(data != fill)
        if factor is not None:
            # make sure the factor is a 64-bit float
            # can't do this in place since data is most likely uint16
            # and we are making it a 64-bit float
            data = data * float(factor) + offset
        data.attrs = attrs

        # handle coordinates (and recursive fun)
        new_coords = {}
        # 'time' dimension causes issues in other processing
        if "time" in data.coords:
            data = data.drop_vars("time")
        if item in data.coords:
            self.coords[item] = data
        for coord_name in data.coords.keys():
            if coord_name not in self.coords:
                self.coords[coord_name] = self[coord_name]
            new_coords[coord_name] = self.coords[coord_name]
        data.coords.update(new_coords)
        return data

    def get_shape(self, key, info):
        """Get the shape of the data."""
        return self.nlines, self.ncols

    def get_dataset(self, key, info):
        """Load a dataset."""
        logger.debug("Reading in get_dataset %s.", key["name"])
        var_name = info.get("file_key", self.filetype_info.get("file_key"))
        if var_name:
            data = self[var_name]
        elif "Sectorized_CMI" in self.nc:
            data = self["Sectorized_CMI"]
        elif "data" in self.nc:
            data = self["data"]
        # NetCDF doesn't support multi-threaded reading, trick it by opening
        # as one whole chunk then split it up before we do any calculations
        data = data.chunk({"x": CHUNK_SIZE, "y": CHUNK_SIZE})

        # convert to satpy standard units
        factor = data.attrs.pop("scale_factor", 1)
        offset = data.attrs.pop("add_offset", 0)
        units = data.attrs.get("units", 1)
        # the '*1' unit is some weird convention added/needed by AWIPS
        if units in ["1", "*1"] and key["calibration"] == "reflectance":
            data *= 100
            factor *= 100  # used for valid_min/max
            data.attrs["units"] = "%"

        # set up all the attributes that might be useful to the user/satpy
        data.attrs.update({"platform_name": self.platform_name,
                           "sensor": data.attrs.get("sensor", self.sensor),
                           })
        if "satellite_longitude" in self.nc.attrs:
            data.attrs["orbital_parameters"] = {
                "projection_longitude": self.nc.attrs["satellite_longitude"],
                "projection_latitude": self.nc.attrs["satellite_latitude"],
                "projection_altitude": self.nc.attrs["satellite_altitude"],
            }

        scene_id = self.nc.attrs.get("scene_id")
        if scene_id is not None:
            data.attrs["scene_id"] = scene_id
        data.attrs.update(key.to_dict())
        data.attrs.pop("_FillValue", None)
        if "valid_min" in data.attrs:
            vmin = data.attrs.pop("valid_min")
            vmax = data.attrs.pop("valid_max")
            vmin = vmin * factor + offset
            vmax = vmax * factor + offset
            data.attrs["valid_min"] = vmin
            data.attrs["valid_max"] = vmax
        return data

    def _get_cf_grid_mapping_var(self):
        """Figure out which grid mapping should be used."""
        gmaps = ["fixedgrid_projection", "goes_imager_projection",
                 "lambert_projection", "polar_projection",
                 "mercator_projection"]
        if "grid_mapping" in self.filename_info:
            gmaps = [self.filename_info.get("grid_mapping")] + gmaps
        for grid_mapping in gmaps:
            if grid_mapping in self.nc:
                return self.nc[grid_mapping]
        raise KeyError("Can't find grid mapping variable in SCMI file")

    def _get_proj4_name(self, projection):
        """Map CF projection name to PROJ.4 name."""
        gmap_name = projection.attrs["grid_mapping_name"]
        proj = {
            "geostationary": "geos",
            "lambert_conformal_conic": "lcc",
            "polar_stereographic": "stere",
            "mercator": "merc",
        }.get(gmap_name, gmap_name)
        return proj

    def _get_proj_specific_params(self, projection):
        """Convert CF projection parameters to PROJ.4 dict."""
        proj = self._get_proj4_name(projection)
        proj_dict = {
            "proj": proj,
            "a": float(projection.attrs["semi_major_axis"]),
            "b": float(projection.attrs["semi_minor_axis"]),
            "units": "m",
        }
        if proj == "geos":
            proj_dict["h"] = float(projection.attrs["perspective_point_height"])
            proj_dict["sweep"] = projection.attrs.get("sweep_angle_axis", "y")
            proj_dict["lon_0"] = float(projection.attrs["longitude_of_projection_origin"])
            proj_dict["lat_0"] = float(projection.attrs.get("latitude_of_projection_origin", 0.0))
        elif proj == "lcc":
            proj_dict["lat_0"] = float(projection.attrs["standard_parallel"])
            proj_dict["lon_0"] = float(projection.attrs["longitude_of_central_meridian"])
            proj_dict["lat_1"] = float(projection.attrs["latitude_of_projection_origin"])
        elif proj == "stere":
            proj_dict["lat_ts"] = float(projection.attrs["standard_parallel"])
            proj_dict["lon_0"] = float(projection.attrs["straight_vertical_longitude_from_pole"])
            proj_dict["lat_0"] = float(projection.attrs["latitude_of_projection_origin"])
        elif proj == "merc":
            proj_dict["lat_ts"] = float(projection.attrs["standard_parallel"])
            proj_dict["lat_0"] = proj_dict["lat_ts"]
            proj_dict["lon_0"] = float(projection.attrs["longitude_of_projection_origin"])
        else:
            raise ValueError("Can't handle projection '{}'".format(proj))
        return proj_dict

    def _calc_extents(self, proj_dict):
        """Calculate area extents from x/y variables."""
        h = float(proj_dict.get("h", 1.))  # force to 64-bit float
        x = self["x"]
        y = self["y"]
        x_units = x.attrs.get("units", "rad")
        if x_units == "meters":
            h_factor = 1.
            factor = 1.
        elif x_units == "microradian":
            h_factor = h
            factor = 1e6
        else:  # radians
            h_factor = h
            factor = 1.
        x_l = h_factor * x[0] / factor
        x_r = h_factor * x[-1] / factor
        y_l = h_factor * y[-1] / factor
        y_u = h_factor * y[0] / factor
        x_half = (x_r - x_l) / (self.ncols - 1) / 2.
        y_half = (y_u - y_l) / (self.nlines - 1) / 2.
        return x_l - x_half, y_l - y_half, x_r + x_half, y_u + y_half

    def get_area_def(self, key):
        """Get the area definition of the data at hand."""
        # FIXME: Can't we pass dataset info to the get_area_def?
        projection = self._get_cf_grid_mapping_var()
        proj_dict = self._get_proj_specific_params(projection)
        area_extent = self._calc_extents(proj_dict)
        area_name = "{}_{}".format(self.sensor, proj_dict["proj"])
        return geometry.AreaDefinition(
            area_name,
            "SCMI file area",
            area_name,
            proj_dict,
            self.ncols,
            self.nlines,
            np.asarray(area_extent))

    @property
    def start_time(self):
        """Get the start time."""
        return dt.datetime.strptime(self.nc.attrs["start_date_time"], "%Y%j%H%M%S")

    @property
    def end_time(self):
        """Get the end time."""
        return self.start_time

    def __del__(self):
        """Delete the instance."""
        try:
            self.nc.close()
        except OSError:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""File handler for DSCOVR EPIC L1B data in hdf5 format.

The ``epic_l1b_h5`` reader reads and calibrates EPIC L1B image data in hdf5 format.

This reader supports all image and most ancillary datasets.
Once the reader is initialised:

`` scn = Scene([epic_filename], reader='epic_l1b_h5')``

Channels can be loaded with the 'B' prefix and their wavelength in nanometers:

``scn.load(['B317', 'B688'])``

while ancillary data can be loaded by its name:

``scn.load(['solar_zenith_angle'])``

Note that ancillary dataset names use common standards and not the dataset names in the file.
By default, channel data is loaded as calibrated reflectances, but counts data is also available.

"""

import datetime as dt
import logging

import dask.array as da
import numpy as np

from satpy.readers.core.hdf5 import HDF5FileHandler

logger = logging.getLogger(__name__)

# Level 1b is given as counts. These factors convert to reflectance.
# Retrieved from: https://asdc.larc.nasa.gov/documents/dscovr/DSCOVR_EPIC_Calibration_Factors_V03.pdf
CALIB_COEFS = {"B317": 1.216e-4,
               "B325": 1.111e-4,
               "B340": 1.975e-5,
               "B388": 2.685e-5,
               "B443": 8.34e-6,
               "B551": 6.66e-6,
               "B680": 9.3e-6,
               "B688": 2.02e-5,
               "B764": 2.36e-5,
               "B780": 1.435e-5}


class DscovrEpicL1BH5FileHandler(HDF5FileHandler):
    """File handler for DSCOVR EPIC L1b data."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init filehandler."""
        super(DscovrEpicL1BH5FileHandler, self).__init__(filename, filename_info, filetype_info)

        self.sensor = "epic"
        self.platform_name = "DSCOVR"

    @property
    def start_time(self):
        """Get the start time."""
        start_time = dt.datetime.strptime(self.file_content["/attr/begin_time"], "%Y-%m-%d %H:%M:%S")
        return start_time

    @property
    def end_time(self):
        """Get the end time."""
        end_time = dt.datetime.strptime(self.file_content["/attr/end_time"], "%Y-%m-%d %H:%M:%S")
        return end_time

    @staticmethod
    def _mask_infinite(band):
        band.data = da.where(np.isfinite(band.data), band.data, np.nan)
        return band

    @staticmethod
    def calibrate(data, ds_name, calibration=None):
        """Convert counts input reflectance."""
        if calibration == "reflectance":
            return data * CALIB_COEFS[ds_name] * 100.
        return data

    def get_dataset(self, dataset_id, ds_info):
        """Load a dataset."""
        ds_name = dataset_id["name"]

        logger.debug("Reading in get_dataset %s.", ds_name)
        file_key = ds_info.get("file_key", ds_name)

        band = self._mask_infinite(self.get(file_key))
        band = self.calibrate(band, ds_name, calibration=dataset_id.get("calibration"))
        band = self._update_metadata(band)

        return band

    def _update_metadata(self, band):
        band = band.rename({band.dims[0]: "x", band.dims[1]: "y"})
        band.attrs.update({"platform_name": self.platform_name, "sensor": self.sensor})

        return band



# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Lookups, colorization and colormaps."""

from __future__ import annotations

import logging
import os

import numpy as np

from satpy._config import get_config_path
from satpy.utils import find_in_ancillary

from .wrappers import exclude_alpha, on_separate_bands, using_map_blocks

LOG = logging.getLogger(__name__)


def lookup(img, **kwargs):
    """Assign values to channels based on a table."""
    luts = np.array(kwargs["luts"], dtype=np.float32) / 255.0
    return _lookup_table(img.data, luts=luts)


@exclude_alpha
@on_separate_bands
@using_map_blocks
def _lookup_table(band_data, luts=None, index=-1):
    # NaN/null values will become 0
    lut = luts[:, index] if len(luts.shape) == 2 else luts
    band_data = band_data.clip(0, lut.size - 1)
    # Convert to uint8, with NaN/null values changed into 0
    band_data = np.nan_to_num(band_data).astype(np.uint8)
    return lut[band_data]


def colorize(img, **kwargs):  # noqa: D417
    """Colorize the given image.

    Args:
        img: image to be colorized
    Kwargs:
        palettes: colormap(s) to use

    The `palettes` kwarg can be one of the following:
        - a trollimage.colormap.Colormap object
        - list of dictionaries with each of one of the following forms:
            - {'filename': '/path/to/colors.npy',
               'min_value': <float, min value to match colors to>,
               'max_value': <float, min value to match colors to>,
               'reverse': <bool, reverse the colormap if True (default: False)}
            - {'colors': <trollimage.colormap.Colormap instance>,
               'min_value': <float, min value to match colors to>,
               'max_value': <float, min value to match colors to>,
               'reverse': <bool, reverse the colormap if True (default: False)}
            - {'colors': <tuple of RGB(A) tuples>,
               'min_value': <float, min value to match colors to>,
               'max_value': <float, min value to match colors to>,
               'reverse': <bool, reverse the colormap if True (default: False)}
            - {'colors': <tuple of RGB(A) tuples>,
               'values': <tuple of values to match colors to>,
               'min_value': <float, min value to match colors to>,
               'max_value': <float, min value to match colors to>,
               'reverse': <bool, reverse the colormap if True (default: False)}
            - {'dataset': <str, referring to dataset containing palette>,
               'color_scale': <int, value to be interpreted as white>,
               'min_value': <float, see above>,
               'max_value': <float, see above>}

    If multiple palettes are supplied, they are concatenated before applied.

    """
    full_cmap = _merge_colormaps(kwargs, img)
    img.colorize(full_cmap)


def palettize(img, **kwargs):
    """Palettize the given image (no color interpolation).

    Arguments as for :func:`colorize`.

    NB: to retain the palette when saving the resulting image, pass
    ``keep_palette=True`` to the save method (either via the Scene class or
    directly in trollimage).
    """
    full_cmap = _merge_colormaps(kwargs, img)
    img.palettize(full_cmap)


def _merge_colormaps(kwargs, img=None):
    """Merge colormaps listed in kwargs."""
    from trollimage.colormap import Colormap
    full_cmap = None

    palette = kwargs["palettes"]
    if isinstance(palette, Colormap):
        full_cmap = palette
    else:
        for itm in palette:
            cmap = create_colormap(itm, img)
            if full_cmap is None:
                full_cmap = cmap
            else:
                full_cmap = full_cmap + cmap

    return full_cmap


def create_colormap(palette, img=None):  # noqa: D417
    """Create colormap of the given numpy file, color vector, or colormap.

    Args:
        palette (dict): Information describing how to create a colormap
            object. See below for more details.

    **From a file**

    Colormaps can be loaded from ``.npy``, ``.npz``, or comma-separated text
    files. Numpy (npy/npz) files should be 2D arrays with rows for each color.
    Comma-separated files should have a row for each color with each column
    representing a single value/channel. The filename to load can be provided
    with the ``filename`` key in the provided palette information. A filename
    ending with ``.npy`` or ``.npz`` is read as a numpy file with
    :func:`numpy.load`. All other extensions are
    read as a comma-separated file. For ``.npz`` files the data must be stored
    as a positional list where the first element represents the colormap to
    use. See :func:`numpy.savez` for more information. The path to the
    colormap can be relative if it is stored in a directory specified by
    :ref:`config_path_setting`. Otherwise it should be an absolute path.

    The colormap is interpreted as 1 of 4 different "colormap modes":
    ``RGB``, ``RGBA``, ``VRGB``, or ``VRGBA``. The
    colormap mode can be forced with the ``colormap_mode`` key in the provided
    palette information. If it is not provided then a default will be chosen
    based on the number of columns in the array (3: RGB, 4: VRGB, 5: VRGBA).

    The "V" in the possible colormap modes represents the control value of
    where that color should be applied. If "V" is not provided in the colormap
    data it defaults to the row index in the colormap array (0, 1, 2, ...)
    divided by the total number of colors to produce a number between 0 and 1.
    See the "Set Range" section below for more information.
    The remaining elements in the colormap array represent the Red (R),
    Green (G), and Blue (B) color to be mapped to.

    See the "Color Scale" section below for more information on the value
    range of provided numbers.

    **From a list**

    Colormaps can be loaded from lists of colors provided by the ``colors``
    key in the provided dictionary. Each element in the list represents a
    single color to be mapped to and can be 3 (RGB) or 4 (RGBA) elements long.
    By default, the value or control point for a color is determined by the
    index in the list (0, 1, 2, ...) divided by the total number of colors
    to produce a number between 0 and 1. This can be overridden by providing a
    ``values`` key in the provided dictionary. See the "Set Range" section
    below for more information.

    See the "Color Scale" section below for more information on the value
    range of provided numbers.

    **From a builtin colormap**

    Colormaps can be loaded by name from the builtin colormaps in the
    ``trollimage``` package. Specify the name with the ``colors``
    key in the provided dictionary (ex. ``{'colors': 'blues'}``).
    See :doc:`trollimage:colormap` for the full list of available colormaps.

    **From an auxiliary variable**

    If the colormap is defined in the same dataset as the data to which the
    colormap shall be applied, this can be indicated with
    ``{'dataset': 'palette_variable'}``, where ``'palette_variable'`` is the
    name of the variable containing the palette.  This variable must be an
    auxiliary variable to the dataset to which the colours are applied.  When
    using this, it is important that one should **not** set ``min_value`` and
    ``max_value`` as those will be taken from the ``valid_range`` attribute
    on the dataset and if those differ from ``min_value`` and ``max_value``,
    the resulting colors will not match the ones in the palette.

    **Color Scale**

    By default colors are expected to be in a 0-255 range. This
    can be overridden by specifying ``color_scale`` in the provided colormap
    information. A common alternative to 255 is ``1`` to specify floating
    point numbers between 0 and 1. The resulting Colormap uses the normalized
    color values (0-1).

    **Set Range**

    By default the control points or values of the Colormap are between 0 and
    1. This means that data values being mapped to a color must also be
    between 0 and 1. When this is not the case, the expected input range of
    the data can be used to configure the Colormap and change the control point
    values. To do this specify the input data range with ``min_value`` and
    ``max_value``. See :meth:`trollimage.colormap.Colormap.set_range` for more
    information.

    **Set Alpha Range**

    The alpha channel of a created colormap can be added and/or modified by
    specifying ``min_alpha`` and ``max_alpha``.
    See :meth:`trollimage.colormap.Colormap.set_alpha_range`  for more info.

    """
    # are colors between 0-255 or 0-1
    color_scale = palette.get("color_scale", 255)
    cmap = _get_cmap_from_palette_info(palette, img, color_scale)

    _reverse_cmap(cmap, palette)
    _set_cmap_range(cmap, palette)
    _set_cmap_alpha_range(cmap, palette, color_scale)

    return cmap


def _reverse_cmap(cmap, palette):
    if palette.get("reverse", False):
        cmap.reverse()


def _set_cmap_range(cmap, palette):
    if "min_value" in palette and "max_value" in palette:
        cmap.set_range(palette["min_value"], palette["max_value"])
    elif "min_value" in palette or "max_value" in palette:
        raise ValueError("Both 'min_value' and 'max_value' must be specified (or neither).")


def _set_cmap_alpha_range(cmap, palette, color_scale):
    if "min_alpha" in palette and "max_alpha" in palette:
        cmap.set_alpha_range(palette["min_alpha"] / color_scale,
                             palette["max_alpha"] / color_scale)
    elif "min_alpha" in palette or "max_alpha" in palette:
        raise ValueError("Both 'min_alpha' and 'max_alpha' must be specified (or neither).")


def _get_cmap_from_palette_info(palette, img, color_scale):
    from trollimage.colormap import Colormap

    fname = palette.get("filename", None)
    colors = palette.get("colors", None)
    dataset = palette.get("dataset", None)
    if fname:
        if not os.path.exists(fname):
            fname = get_config_path(fname)
        cmap = Colormap.from_file(fname, palette.get("colormap_mode", None), color_scale)
    elif isinstance(colors, (tuple, list)):
        cmap = Colormap.from_sequence_of_colors(colors, palette.get("values", None), color_scale)
    elif isinstance(colors, str):
        cmap = Colormap.from_name(colors)
    elif isinstance(dataset, str):
        cmap = _create_colormap_from_dataset(img, dataset, color_scale)
    else:
        raise ValueError("Unknown colormap format: {}".format(palette))
    return cmap



def _create_colormap_from_dataset(img, dataset, color_scale):
    """Create a colormap from an auxiliary variable in a source file."""
    from trollimage.colormap import Colormap

    match = find_in_ancillary(img.data, dataset)
    return Colormap.from_array_with_metadata(
        match, img.data.dtype, color_scale,
        valid_range=img.data.attrs.get("valid_range"),
        scale_factor=img.data.attrs.get("scale_factor", 1),
        add_offset=img.data.attrs.get("add_offset", 0),
        remove_last=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for the management of VII products."""


# PLANCK COEFFICIENTS FOR CALIBRATION AS DEFINED BY EUMETSAT
C1 = 1.191062e+8   # [W/m2·sr-1·µm4]
C2 = 1.4387863e+4  # [K·µm]

# CONSTANTS DEFINING THE TIE POINTS
TIE_POINTS_FACTOR = 8    # Sub-sampling factor of tie points wrt pixel points
SCAN_ALT_TIE_POINTS = 4  # Number of tie points along the satellite track for each scan

# MEAN EARTH RADIUS AS DEFINED BY IUGG
MEAN_EARTH_RADIUS = 6371008.7714  # [m]



#!/usr/bin/python
# Copyright (c) 2015 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test objects in writers init module."""

from __future__ import annotations

import pytest


@pytest.mark.parametrize(
    "name",
    [
        "read_writer_config",
        "load_writer_configs",
        "load_writer",
        "configs_for_writer",
        "available_writers",
        "add_overlay",
        "add_text",
        "add_logo",
        "add_scale",
        "add_decorate",
        "get_enhanced_image",
        "split_results",
        "group_results_by_output_file",
        "compute_writer_results",
        "Writer",
        "ImageWriter",
    ],
)
def test_deprecated_imports(name: str):
    """Test that moved objects can be imported but warn."""
    import importlib

    writers_mod = importlib.import_module("satpy.writers")
    with pytest.warns(UserWarning, match=".*has been moved.*") as warn_catcher:
        old_obj = getattr(writers_mod, name)

    assert len(warn_catcher) == 1
    w = warn_catcher[0]
    new_imp = str(w.message).split("'")[3]
    new_mod = importlib.import_module(new_imp.rsplit(".", 1)[0])
    new_obj = getattr(new_mod, name)
    assert new_obj is old_obj



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit testing the stretching enhancements functions."""

import numpy as np
import pytest

from .utils import create_ch1, create_ch2, create_rgb, run_and_check_enhancement, run_and_check_enhancement_with_dtype


class TestEnhancementsStretching:
    """Class for testing enhancements in satpy.enhancements.contrast module."""

    def setup_method(self):
        """Create test data used by every test."""
        self.ch1 = create_ch1()
        self.ch2 = create_ch2()
        self.rgb = create_rgb()

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    def test_cira_stretch(self, dtype):
        """Test applying the cira_stretch."""
        from satpy.enhancements.contrast import cira_stretch

        expected = np.array([[
            [np.nan, -7.04045974, -7.04045974, 0.79630132, 0.95947296],
            [1.05181359, 1.11651012, 1.16635571, 1.20691137, 1.24110186]]], dtype=dtype)
        run_and_check_enhancement_with_dtype(cira_stretch, self.ch1.astype(dtype), expected)

    def test_reinhard(self):
        """Test the reinhard algorithm."""
        from satpy.enhancements.contrast import reinhard_to_srgb
        expected = np.array([[[np.nan, 0., 0., 0.93333793, 1.29432402],
                              [1.55428709, 1.76572249, 1.94738635, 2.10848544, 2.25432809]],

                             [[np.nan, 0., 0., 0.93333793, 1.29432402],
                              [1.55428709, 1.76572249, 1.94738635, 2.10848544, 2.25432809]],

                             [[np.nan, 0., 0., 0.93333793, 1.29432402],
                              [1.55428709, 1.76572249, 1.94738635, 2.10848544, 2.25432809]]])
        run_and_check_enhancement(reinhard_to_srgb, self.rgb, expected)

    def test_piecewise_linear_stretch(self):
        """Test the piecewise_linear_stretch enhancement function."""
        from satpy.enhancements.contrast import piecewise_linear_stretch
        expected = np.array([[
            [np.nan, 0., 0., 0.44378, 0.631734],
            [0.737562, 0.825041, 0.912521, 1., 1.]]])
        run_and_check_enhancement(piecewise_linear_stretch,
                                  self.ch2 / 100.0,
                                  expected,
                                  xp=[0., 25., 55., 100., 255.],
                                  fp=[0., 90., 140., 175., 255.],
                                  reference_scale_factor=255,
                                  )

    def test_btemp_threshold(self):
        """Test applying the cira_stretch."""
        from satpy.enhancements.contrast import btemp_threshold

        expected = np.array([[
            [np.nan, 0.946207, 0.892695, 0.839184, 0.785672],
            [0.73216, 0.595869, 0.158745, -0.278379, -0.715503]]])
        run_and_check_enhancement(btemp_threshold, self.ch1, expected,
                                  min_in=-200, max_in=500, threshold=350)

    def tearDown(self):
        """Clean up."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2023 Satpy Developers


# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test common VIIRS/ATMS SDR reader functions."""

import logging

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.readers.core.viirs_atms_sdr import _get_file_units, _get_scale_factors_for_units
from satpy.tests.utils import make_dataid

DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)


def test_get_file_units(caplog):
    """Test get the file-units from the dataset info."""
    did = make_dataid(name="some_variable", modifiers=())
    ds_info = {"file_units": None}
    with caplog.at_level(logging.DEBUG):
        file_units = _get_file_units(did, ds_info)

    assert file_units is None
    log_output = "Unknown units for file key 'DataID(name='some_variable', modifiers=())'"
    assert log_output in caplog.text


def test_get_scale_factors_for_units_unsupported_units():
    """Test get scale factors for units, when units are not supported."""
    factors = xr.DataArray(da.from_array(DEFAULT_FILE_FACTORS, chunks=1))
    file_units = "unknown unit"
    output_units = "%"
    with pytest.raises(ValueError, match="Don't know how to convert 'unknown unit' to '%'"):
        _ = _get_scale_factors_for_units(factors, file_units, output_units)


def test_get_scale_factors_for_units_reflectances(caplog):
    """Test get scale factors for units, when variable is supposed to be a reflectance."""
    factors = xr.DataArray(da.from_array(DEFAULT_FILE_FACTORS, chunks=1))
    file_units = "1"
    output_units = "%"
    with caplog.at_level(logging.DEBUG):
        retv = _get_scale_factors_for_units(factors, file_units, output_units)

    log_output = "Adjusting scaling factors to convert '1' to '%'"
    assert log_output in caplog.text
    np.testing.assert_allclose(retv, np.array([200., 100.]))


def test_get_scale_factors_for_units_tbs(caplog):
    """Test get scale factors for units, when variable is supposed to be a brightness temperature."""
    factors = xr.DataArray(da.from_array(DEFAULT_FILE_FACTORS, chunks=1))
    file_units = "W cm-2 sr-1"
    output_units = "W m-2 sr-1"
    with caplog.at_level(logging.DEBUG):
        retv = _get_scale_factors_for_units(factors, file_units, output_units)

    log_output = "Adjusting scaling factors to convert 'W cm-2 sr-1' to 'W m-2 sr-1'"
    assert log_output in caplog.text
    np.testing.assert_allclose(retv, np.array([20000., 10000.]))



# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Writer for netCDF4/CF.

Example usage
-------------

The CF writer saves datasets in a Scene as `CF-compliant`_ netCDF file. Here is an example with MSG SEVIRI data in HRIT
format:

    >>> from satpy import Scene
    >>> import glob
    >>> filenames = glob.glob('data/H*201903011200*')
    >>> scn = Scene(filenames=filenames, reader='seviri_l1b_hrit')
    >>> scn.load(['VIS006', 'IR_108'])
    >>> scn.save_datasets(writer='cf', datasets=['VIS006', 'IR_108'], filename='seviri_test.nc',
                          exclude_attrs=['raw_metadata'])

* You can select the netCDF backend using the ``engine`` keyword argument. If `None` if follows
  :meth:`~xarray.Dataset.to_netcdf` engine choices with a preference for 'netcdf4'.
* For datasets with area definition you can exclude lat/lon coordinates by setting ``include_lonlats=False``.
  If the area has a projected CRS, units are assumed to be in metre.  If the
  area has a geographic CRS, units are assumed to be in degrees.  The writer
  does not verify that the CRS is supported by the CF conventions.  One
  commonly used projected CRS not supported by the CF conventions is the
  equirectangular projection, such as EPSG 4087.
* By default non-dimensional coordinates (such as scanline timestamps) are prefixed with the corresponding
  dataset name. This is because they are likely to be different for each dataset. If a non-dimensional
  coordinate is identical for all datasets, the prefix can be removed by setting ``pretty=True``.
* Some dataset names start with a digit, like AVHRR channels 1, 2, 3a, 3b, 4 and 5. This doesn't comply with CF
  https://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/build/ch02s03.html. These channels are prefixed
  with ``"CHANNEL_"`` by default. This can be controlled with the variable `numeric_name_prefix` to `save_datasets`.
  Setting it to `None` or `''` will skip the prefixing.

Grouping
~~~~~~~~

All datasets to be saved must have the same projection coordinates ``x`` and ``y``. If a scene holds datasets with
different grids, the CF compliant workaround is to save the datasets to separate files. Alternatively, you can save
datasets with common grids in separate netCDF groups as follows:

    >>> scn.load(['VIS006', 'IR_108', 'HRV'])
    >>> scn.save_datasets(writer='cf', datasets=['VIS006', 'IR_108', 'HRV'],
                          filename='seviri_test.nc', exclude_attrs=['raw_metadata'],
                          groups={'visir': ['VIS006', 'IR_108'], 'hrv': ['HRV']})

Note that the resulting file will not be fully CF compliant.


Dataset Encoding
~~~~~~~~~~~~~~~~

Dataset encoding can be specified in two ways:

1) Via the ``encoding`` keyword argument of ``save_datasets``:

    >>> my_encoding = {
    ...    'my_dataset_1': {
    ...        'compression': 'zlib',
    ...        'complevel': 9,
    ...        'scale_factor': 0.01,
    ...        'add_offset': 100,
    ...        'dtype': np.int16
    ...     },
    ...    'my_dataset_2': {
    ...        'compression': None,
    ...        'dtype': np.float64
    ...     }
    ... }
    >>> scn.save_datasets(writer='cf', filename='encoding_test.nc', encoding=my_encoding)


2) Via the ``encoding`` attribute of the datasets in a scene. For example

    >>> scn['my_dataset'].encoding = {'compression': 'zlib'}
    >>> scn.save_datasets(writer='cf', filename='encoding_test.nc')

See the `xarray encoding documentation`_ for all encoding options.

.. note::

    Chunk-based compression can be specified with the ``compression`` keyword
    since

        .. code::

            netCDF4-1.6.0
            libnetcdf-4.9.0
            xarray-2022.12.0

    The ``zlib`` keyword is deprecated. Make sure that the versions of
    these modules are all above or all below that reference. Otherwise,
    compression might fail or be ignored silently.


Attribute Encoding
~~~~~~~~~~~~~~~~~~

In the above examples, raw metadata from the HRIT files have been excluded. If you want all attributes to be included,
just remove the ``exclude_attrs`` keyword argument. By default, dict-type dataset attributes, such as the raw metadata,
are encoded as a string using json. Thus, you can use json to decode them afterwards:

    >>> import xarray as xr
    >>> import json
    >>> # Save scene to nc-file
    >>> scn.save_datasets(writer='cf', datasets=['VIS006', 'IR_108'], filename='seviri_test.nc')
    >>> # Now read data from the nc-file
    >>> ds = xr.open_dataset('seviri_test.nc')
    >>> raw_mda = json.loads(ds['IR_108'].attrs['raw_metadata'])
    >>> print(raw_mda['RadiometricProcessing']['Level15ImageCalibration']['CalSlope'])
    [0.020865   0.0278287  0.0232411  0.00365867 0.00831811 0.03862197
     0.12674432 0.10396091 0.20503568 0.22231115 0.1576069  0.0352385]


Alternatively it is possible to flatten dict-type attributes by setting ``flatten_attrs=True``. This is more human
readable as it will create a separate nc-attribute for each item in every dictionary. Keys are concatenated with
underscore separators. The `CalSlope` attribute can then be accessed as follows:

    >>> scn.save_datasets(writer='cf', datasets=['VIS006', 'IR_108'], filename='seviri_test.nc',
                          flatten_attrs=True)
    >>> ds = xr.open_dataset('seviri_test.nc')
    >>> print(ds['IR_108'].attrs['raw_metadata_RadiometricProcessing_Level15ImageCalibration_CalSlope'])
    [0.020865   0.0278287  0.0232411  0.00365867 0.00831811 0.03862197
     0.12674432 0.10396091 0.20503568 0.22231115 0.1576069  0.0352385]

This is what the corresponding ``ncdump`` output would look like in this case:

.. code-block:: none

    $ ncdump -h test_seviri.nc
    ...
    IR_108:raw_metadata_RadiometricProcessing_Level15ImageCalibration_CalOffset = -1.064, ...;
    IR_108:raw_metadata_RadiometricProcessing_Level15ImageCalibration_CalSlope = 0.021, ...;
    IR_108:raw_metadata_RadiometricProcessing_MPEFCalFeedback_AbsCalCoeff = 0.021, ...;
    ...


.. _CF-compliant: http://cfconventions.org/
.. _xarray encoding documentation:
    http://xarray.pydata.org/en/stable/user-guide/io.html?highlight=encoding#writing-encoded-data
"""
import copy
import logging
import warnings

import numpy as np
import xarray as xr
from packaging.version import InvalidVersion, Version

from satpy.cf.coords import EPOCH  # noqa: F401 (for backward compatibility)
from satpy.writers.core.base import Writer

logger = logging.getLogger(__name__)

# Check availability of either netCDF4 or h5netcdf package
try:
    import netCDF4
except ImportError:
    netCDF4 = None

try:
    import h5netcdf
except ImportError:
    h5netcdf = None

# Ensure that either netCDF4 or h5netcdf is available to avoid silent failure
if netCDF4 is None and h5netcdf is None:
    raise ImportError("Ensure that the netCDF4 or h5netcdf package is installed.")


CF_VERSION = "CF-1.7"


# Numpy datatypes compatible with all netCDF4 backends. ``np.str_`` is
# excluded because h5py (and thus h5netcdf) has problems with unicode, see
# https://github.com/h5py/h5py/issues/624."""
NC4_DTYPES = [np.dtype("int8"), np.dtype("uint8"),
              np.dtype("int16"), np.dtype("uint16"),
              np.dtype("int32"), np.dtype("uint32"),
              np.dtype("int64"), np.dtype("uint64"),
              np.dtype("float32"), np.dtype("float64"),
              np.bytes_]

# Unsigned and int64 isn't CF 1.7 compatible
# Note: Unsigned and int64 are CF 1.9 compatible
CF_DTYPES = [np.dtype("int8"),
             np.dtype("int16"),
             np.dtype("int32"),
             np.dtype("float32"),
             np.dtype("float64"),
             np.bytes_]


def _sanitize_writer_kwargs(writer_kwargs):
    """Remove satpy-specific kwargs."""
    writer_kwargs = copy.deepcopy(writer_kwargs)
    satpy_kwargs = ["overlay", "decorate", "config_files"]
    for kwarg in satpy_kwargs:
        writer_kwargs.pop(kwarg, None)
    return writer_kwargs


def _initialize_root_netcdf(filename, engine, header_attrs, to_netcdf_kwargs):
    """Initialize root empty netCDF."""
    root = xr.Dataset({}, attrs=header_attrs)
    init_nc_kwargs = to_netcdf_kwargs.copy()
    init_nc_kwargs.pop("encoding", None)  # No variables to be encoded at this point
    init_nc_kwargs.pop("unlimited_dims", None)
    written = [root.to_netcdf(filename, engine=engine, mode="w", **init_nc_kwargs)]
    return written


class CFWriter(Writer):
    """Writer producing NetCDF/CF compatible datasets."""

    def save_dataset(self, dataset, filename=None, fill_value=None, **kwargs):
        """Save the *dataset* to a given *filename*."""
        return self.save_datasets([dataset], filename, **kwargs)

    def save_datasets(self, datasets, filename=None, groups=None, header_attrs=None, engine=None, epoch=None,  # noqa: D417
                      flatten_attrs=False, exclude_attrs=None, include_lonlats=True, pretty=False,
                      include_orig_name=True, numeric_name_prefix="CHANNEL_", **to_netcdf_kwargs):
        """Save the given datasets in one netCDF file.

        Note that all datasets (if grouping: in one group) must have the same projection coordinates.

        Args:
            datasets (list): List of xr.DataArray to be saved.
            filename (str): Output file.
            groups (dict): Group datasets according to the given assignment:
                `{'group_name': ['dataset1', 'dataset2', ...]}`.
                The group name `None` corresponds to the root of the file, i.e., no group will be created.
                Warning: The results will not be fully CF compliant!
            header_attrs: Global attributes to be included.
            engine (str, Optional): Module to be used for writing netCDF files. Follows xarray's
                :meth:`~xarray.Dataset.to_netcdf` engine choices with a preference for 'netcdf4'.
            epoch (str, Optional): Reference time for encoding of time coordinates.
                If None, the default reference time is defined using `from satpy.cf.coords import EPOCH`.
            flatten_attrs (bool, Optional): If True, flatten dict-type attributes.
            exclude_attrs (list, Optional): List of dataset attributes to be excluded.
            include_lonlats (bool, Optional): Always include latitude and longitude coordinates,
                even for datasets with area definition.
            pretty (bool, Optional): Don't modify coordinate names, if possible.
                Makes the file prettier, but possibly less consistent.
            include_orig_name (bool, Optional): Include the original dataset name as a variable
                attribute in the final netCDF.
            numeric_name_prefix (str, Optional): Prefix to add to each variable with a name starting with a digit.
                Use '' or None to leave this out.
        """
        from satpy.cf.datasets import collect_cf_datasets
        from satpy.cf.encoding import update_encoding

        logger.info("Saving datasets to NetCDF4/CF.")
        _check_backend_versions()

        # Define netCDF filename if not provided
        # - It infers the name from the first DataArray
        filename = filename or self.get_filename(**datasets[0].attrs)

        # Collect xr.Dataset for each group
        grouped_datasets, header_attrs = collect_cf_datasets(list_dataarrays=datasets,  # list of xr.DataArray
                                                             header_attrs=header_attrs,
                                                             exclude_attrs=exclude_attrs,
                                                             flatten_attrs=flatten_attrs,
                                                             pretty=pretty,
                                                             include_lonlats=include_lonlats,
                                                             epoch=epoch,
                                                             include_orig_name=include_orig_name,
                                                             numeric_name_prefix=numeric_name_prefix,
                                                             groups=groups,
                                                             )
        # Remove satpy-specific kwargs
        # - This kwargs can contain encoding dictionary
        to_netcdf_kwargs = _sanitize_writer_kwargs(to_netcdf_kwargs)

        # If writing grouped netCDF, create an empty "root" netCDF file
        # - Add the global attributes
        # - All groups will be appended in the for loop below
        if groups is not None:
            written = _initialize_root_netcdf(filename=filename,
                                              engine=engine,
                                              header_attrs=header_attrs,
                                              to_netcdf_kwargs=to_netcdf_kwargs)
            mode = "a"
        else:
            mode = "w"
            written = []

        # Write the netCDF
        # - If grouped netCDF, it appends to the root file
        # - If single netCDF, it write directly
        for group_name, ds in grouped_datasets.items():
            encoding, other_to_netcdf_kwargs = update_encoding(ds,
                                                               to_engine_kwargs=to_netcdf_kwargs,
                                                               numeric_name_prefix=numeric_name_prefix)
            res = ds.to_netcdf(filename,
                               engine=engine,
                               group=group_name,
                               mode=mode,
                               encoding=encoding,
                               **other_to_netcdf_kwargs)
            written.append(res)
        return written

    @staticmethod
    def da2cf(dataarray, epoch=None, flatten_attrs=False, exclude_attrs=None,
              include_orig_name=True, numeric_name_prefix="CHANNEL_"):
        """Convert the dataarray to something cf-compatible.

        Args:
            dataarray (xarray.DataArray):
                The data array to be converted.
            epoch (str):
                Reference time for encoding of time coordinates.
                If None, the default reference time is defined using `from satpy.cf.coords import EPOCH`
            flatten_attrs (bool):
                If True, flatten dict-type attributes.
            exclude_attrs (list):
                List of dataset attributes to be excluded.
            include_orig_name (bool):
                Include the original dataset name in the netcdf variable attributes.
            numeric_name_prefix (str):
                Prepend dataset name with this if starting with a digit.
        """
        from satpy.cf.data_array import make_cf_data_array
        warnings.warn("CFWriter.da2cf is deprecated."
                      "Use satpy.cf.dataarray.make_cf_data_array instead.",
                      DeprecationWarning, stacklevel=3)
        return make_cf_data_array(dataarray=dataarray,
                                  epoch=epoch,
                                  flatten_attrs=flatten_attrs,
                                  exclude_attrs=exclude_attrs,
                                  include_orig_name=include_orig_name,
                                  numeric_name_prefix=numeric_name_prefix)

    @staticmethod
    def update_encoding(dataset, to_netcdf_kwargs):
        """Update encoding info (deprecated)."""
        from satpy.cf.encoding import update_encoding

        warnings.warn("CFWriter.update_encoding is deprecated. "
                      "Use satpy.cf.encoding.update_encoding instead.",
                      DeprecationWarning, stacklevel=3)
        return update_encoding(dataset, to_netcdf_kwargs)


# --------------------------------------------------------------------------.
# NetCDF version


def _check_backend_versions():
    """Issue warning if backend versions do not match."""
    if not _backend_versions_match():
        warnings.warn(
            "Backend version mismatch. Compression might fail or be ignored "
            "silently. Recommended: All versions below or above "
            "netCDF4-1.6.0/libnetcdf-4.9.0/xarray-2022.12.0.",
            stacklevel=3
        )


def _backend_versions_match():
    versions = _get_backend_versions()
    reference = {
        "netCDF4": Version("1.6.0"),
        "libnetcdf": Version("4.9.0"),
        "xarray": Version("2022.12.0")
    }
    is_newer = [
        versions[module] >= reference[module]
        for module in versions
    ]
    all_newer = all(is_newer)
    all_older = not any(is_newer)
    return all_newer or all_older


def _get_backend_versions():
    import netCDF4
    libnetcdf_version = _parse_libnetcdf_version(
        netCDF4.__netcdf4libversion__
    )
    return {
        "netCDF4": Version(netCDF4.__version__),
        "libnetcdf": libnetcdf_version,
        "xarray": Version(xr.__version__)
    }


def _parse_libnetcdf_version(version_str):
    # Make libnetcdf development version compatible with PEP440
    version_str = version_str.replace("development", "dev")
    try:
        return Version(version_str)
    except InvalidVersion:
        warnings.warn(
            f"Unable to parse netcdf-c version {version_str}, "
            f"using 0.0.0 as fallback",
            UserWarning,
            stacklevel=3
        )
        return Version("0.0.0")
